# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2018  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Markdown support"""

# pylint: disable=W0702
# pylint: disable=W0703

import os.path
import mistune
from pygments import highlight
from pygments.lexers import get_lexer_by_name, get_lexer_for_mimetype
from pygments.formatters import HtmlFormatter
from utils.fileutils import getMagicMimeFromBuffer
from .settings import Settings


# Notes:
# QTextBrowser is used to display the generated HTML. Unfortunately,
# QTextBrowser supports only a very specific set of attributes and sometimes
# renders the content quite unexpectedly. In particular, <div ...> are not
# really supported. So, styling of the markdown code blocks and quotes are done
# via tables.
# QWebEngineView (which supposedly is better in terms of rendering) does not
# work on all platforms out of the box (at least when used via PyQt).
# See also: http://www.prog.org.ru/topic_7398_0.html


CODE_BLOCK_STYLE = 'style="border-width:1px;border-style:solid;' \
                   'border-color:#ccc;margin-top:.5em;margin-bottom:.5em;"'
PRE_WRAP_START = '''<table cellspacing="0" cellpadding="8" width="100%"
                           align="left" bgcolor="#f8f8f8"''' + \
                 CODE_BLOCK_STYLE + '><tr><td>'
PRE_WRAP_END = '</td></tr></table>'

QUOTE_TABLE = '<table cellspacing="0" width="100%" align="left" ' \
              'style="margin-top:.4em;margin-bottom:.4em;">'


def get_lexer(text, lang):
    """Tries to get lexer for the text whether the lang is provided or not"""
    if lang:
        try:
            return get_lexer_by_name(lang, stripall=False)
        except:
            pass
        return None

    # No language provided, try to guess
    mime = getMagicMimeFromBuffer(text.strip())
    if mime:
        try:
            return get_lexer_for_mimetype(mime, stripall=False)
        except:
            pass

        # The pygments data sometimes miss mime options provided by python magic
        # library
        if mime.startswith('text/'):
            try:
                return get_lexer_for_mimetype(mime.replace('text/',
                                                           'application/'),
                                              stripall=False)
            except:
                pass

    return None


# The start/end tags could be:
# @startuml / @enduml
# @startgantt / @endgantt
# @startsalt / @endsalt
# @startmindmap / @endmindmap
# @startwbs / @endwbs
# @startditaa / @endditaa
# @startjcckit / @endjcckit
def is_plant_uml(text, lang):
    """True if it is a plant uml diagram"""
    if lang:
        return lang.lower() in ['plantuml', 'uml']
    return text.lstrip().lower().startswith('@start')


def block_code(uuid, text, lang, inlinestyles=False, linenos=False):
    """Renders a code block"""
    if is_plant_uml(text, lang):
        fName = Settings().plantUMLCache.getResource(text, uuid)
        if fName is None:
            return '<br/><img src="cdm-image:na50.png"><br/>\n'
        return '<br/><img src="plantuml:' + fName + '"><br/>\n'

    lexer = get_lexer(text, lang)
    if lexer:
        try:
            formatter = HtmlFormatter(noclasses=inlinestyles, linenos=linenos)
            code = highlight(text, lexer, formatter)
            return ''.join([PRE_WRAP_START, '<pre>',
                            code.replace('125%', '100%'), '</pre>',
                            PRE_WRAP_END, '\n'])
        except:
            pass

    return ''.join([PRE_WRAP_START, '<pre>', mistune.escape(text),
                    '</pre>', PRE_WRAP_END, '\n'])


class CDMMarkdownRenderer(mistune.Renderer):

    """Codimension custom markdown renderer"""

    def __init__(self, uuid, fileName):
        mistune.Renderer.__init__(self, inlinestyles=True, linenos=False)
        self.__uuid = uuid
        self.__fileName = fileName

    def block_code(self, code, lang=None):
        """Custom block code renderer"""
        # renderer has an options
        inlinestyles = self.options.get('inlinestyles', False)
        linenos = self.options.get('linenos', False)
        return block_code(self.__uuid, code, lang, inlinestyles, linenos)

    def block_quote(self, text):
        """Custom block quote renderer"""
        # The text comes as \n separated paragraphs
        return ''.join(['<p>', QUOTE_TABLE,
                        '<tr><td bgcolor="#ccc" width="4"></td>',
                        '<td width="8"></td>',
                        '<td>',
                        text.rstrip('\n')
                        .replace('</p>\n', '</p>').replace('\n', '<br/>'),
                        '</td></tr></table></p>'])

    def image(self, src, title, text):
        """Custom image handler"""
        if src and self.__fileName:
            if not os.path.isabs(src):
                newSrcPath = ''.join([os.path.dirname(self.__fileName),
                                      os.path.sep, src])
                src = os.path.normpath(newSrcPath)
        return mistune.Renderer.image(self, src, title, text)

    def codespan(self, text):
        """Custom code span renderer"""
        return '<u>' + mistune.Renderer.codespan(self, text) + '</u>'

    def table(self, header, body):
        """Custom table tag renderer"""
        replacement = '<table cellspacing="0" cellpadding="4"' + \
                      CODE_BLOCK_STYLE + '>'
        return mistune.Renderer.table(self, header, body).replace('<table>',
                                                                  replacement)



def renderMarkdown(uuid, text, fileName):
    """Renders the given text"""
    warnings = []
    errors = []
    renderedText = None
    try:
        renderer = CDMMarkdownRenderer(uuid, fileName)
        markdown = mistune.Markdown(renderer=renderer)
        renderedText = markdown(text)
    except Exception as exc:
        errors.append(str(exc))
    except:
        errors.append('Unknown markdown rendering exception')
    return renderedText, errors, warnings



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Profiling results widget"""

import pstats
from ui.qt import (Qt, QSize, pyqtSignal, QWidget, QToolBar, QHBoxLayout,
                   QAction)
from ui.mainwindowtabwidgetbase import MainWindowTabWidgetBase
from utils.pixmapcache import getIcon
from .proftable import ProfileTableViewer
from .profgraph import ProfileGraphViewer


class ProfileResultsWidget(QWidget, MainWindowTabWidgetBase):

    """Profiling results widget"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, scriptName, params, reportTime, dataFile, parent=None):
        MainWindowTabWidgetBase.__init__(self)
        QWidget.__init__(self, parent)

        # The same stats object is needed for both - a table and a graph
        # So, parse profile output once and then pass the object further
        stats = pstats.Stats(dataFile)
        stats.calc_callees()

        self.__profTable = ProfileTableViewer(scriptName, params, reportTime,
                                              dataFile, stats, self)
        self.__profGraph = ProfileGraphViewer(scriptName, params, reportTime,
                                              dataFile, stats, self)
        self.__profTable.hide()

        self.__profTable.sigEscapePressed.connect(self.__onEsc)
        self.__profGraph.sigEscapePressed.connect(self.__onEsc)

        self.__createLayout()

    def __createLayout(self):
        """Creates the toolbar and layout"""
        # Buttons
        self.__toggleViewButton = QAction(getIcon('tableview.png'),
                                          'Switch to table view', self)
        self.__toggleViewButton.setCheckable(True)
        self.__toggleViewButton.toggled.connect(self.__switchView)

        self.__togglePathButton = QAction(getIcon('longpath.png'),
                                          'Show full paths for item location',
                                          self)
        self.__togglePathButton.setCheckable(True)
        self.__togglePathButton.toggled.connect(self.__togglePath)
        self.__togglePathButton.setEnabled(False)

        self.__printButton = QAction(getIcon('printer.png'), 'Print', self)
        self.__printButton.triggered.connect(self.__onPrint)
        self.__printButton.setEnabled(False)

        self.__printPreviewButton = QAction(getIcon('printpreview.png'),
                                            'Print preview', self)
        self.__printPreviewButton.triggered.connect(self.__onPrintPreview)
        self.__printPreviewButton.setEnabled(False)

        fixedSpacer = QWidget()
        fixedSpacer.setFixedHeight(16)

        self.__zoomInButton = QAction(getIcon('zoomin.png'),
                                      'Zoom in (Ctrl+=)', self)
        self.__zoomInButton.setShortcut('Ctrl+=')
        self.__zoomInButton.triggered.connect(self.onZoomIn)

        self.__zoomOutButton = QAction(getIcon('zoomout.png'),
                                       'Zoom out (Ctrl+-)', self)
        self.__zoomOutButton.setShortcut('Ctrl+-')
        self.__zoomOutButton.triggered.connect(self.onZoomOut)

        self.__zoomResetButton = QAction(getIcon('zoomreset.png'),
                                         'Zoom reset (Ctrl+0)', self)
        self.__zoomResetButton.setShortcut('Ctrl+0')
        self.__zoomResetButton.triggered.connect(self.onZoomReset)

        # Toolbar
        toolbar = QToolBar(self)
        toolbar.setOrientation(Qt.Vertical)
        toolbar.setMovable(False)
        toolbar.setAllowedAreas(Qt.RightToolBarArea)
        toolbar.setIconSize(QSize(16, 16))
        toolbar.setFixedWidth(28)
        toolbar.setContentsMargins(0, 0, 0, 0)

        toolbar.addAction(self.__toggleViewButton)
        toolbar.addAction(self.__togglePathButton)
        toolbar.addAction(self.__printPreviewButton)
        toolbar.addAction(self.__printButton)
        toolbar.addWidget(fixedSpacer)
        toolbar.addAction(self.__zoomInButton)
        toolbar.addAction(self.__zoomOutButton)
        toolbar.addAction(self.__zoomResetButton)

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)
        hLayout.addWidget(self.__profTable)
        hLayout.addWidget(self.__profGraph)
        hLayout.addWidget(toolbar)

        self.setLayout(hLayout)

    def setFocus(self):
        """Overriden setFocus"""
        if self.__profTable.isVisible():
            self.__profTable.setFocus()
        else:
            self.__profGraph.setFocus()

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def __switchView(self, state):
        """Triggered when view is to be switched"""
        if state:
            self.__profGraph.hide()
            self.__profTable.show()
            self.__toggleViewButton.setIcon(getIcon('profdgmview.png'))
            self.__toggleViewButton.setToolTip('Switch to diagram view')
            self.__zoomInButton.setEnabled(False)
            self.__zoomOutButton.setEnabled(False)
            self.__zoomResetButton.setEnabled(False)
            self.__togglePathButton.setEnabled(True)
            self.__profTable.setFocus()
        else:
            self.__profTable.hide()
            self.__profGraph.show()
            self.__toggleViewButton.setIcon(getIcon('tableview.png'))
            self.__toggleViewButton.setToolTip('Switch to table view')
            self.__zoomInButton.setEnabled(True)
            self.__zoomOutButton.setEnabled(True)
            self.__zoomResetButton.setEnabled(True)
            self.__togglePathButton.setEnabled(False)
            self.__profGraph.setFocus()

    def __togglePath(self, state):
        """Triggered when full path/file name is switched"""
        self.__profTable.togglePath(state)
        if state:
            fName = 'shortpath.png'
            tip = 'Show file names only for item location'
        else:
            fName = 'longpath.png'
            tip = 'Show full paths for item location'
        self.__togglePathButton.setIcon(getIcon(fName))
        self.__togglePathButton.setToolTip(tip)

    def __onPrint(self):
        """Triggered on the 'print' button"""
        pass

    def __onPrintPreview(self):
        """Triggered on the 'print preview' button"""
        pass

    def isZoomApplicable(self):
        """Should the zoom menu items be available"""
        return self.__profGraph.isVisible()

    def onZoomIn(self):
        """Triggered on the 'zoom in' button"""
        if self.__profGraph.isVisible():
            self.__profGraph.zoomIn()

    def onZoomOut(self):
        """Triggered on the 'zoom out' button"""
        if self.__profGraph.isVisible():
            self.__profGraph.zoomOut()

    def onZoomReset(self):
        """Triggered on the 'zoom reset' button"""
        if self.__profGraph.isVisible():
            self.__profGraph.resetZoom()

    def isCopyAvailable(self):
        """Tells id the main menu copy item should be switched on"""
        return self.__profGraph.isVisible()

    def isDiagramActive(self):
        """Tells if the diagram is active"""
        return self.__profGraph.isVisible()

    def onCopy(self):
        """Ctrl+C triggered"""
        if self.__profGraph.isVisible():
            self.__profGraph.onCopy()

    def onSaveAs(self, fileName):
        """Saves the diagram into a file"""
        if self.__profGraph.isVisible():
            self.__profGraph.onSaveAs(fileName)
        else:
            self.__profTable.onSaveAs(fileName)

    # Mandatory interface part is below

    def isModified(self):
        """Tells if the file is modified"""
        return False

    def getRWMode(self):
        """Tells if the file is read only"""
        return "RO"

    def getType(self):
        """Tells the widget type"""
        return MainWindowTabWidgetBase.ProfileViewer

    def getLanguage(self):
        """Tells the content language"""
        return "Profiler"

    def setFileName(self, name):
        """Sets the file name - not applicable"""
        raise Exception("Setting a file name for profile results "
                        "is not applicable")

    def setEncoding(self, newEncoding):
        """Not applicable for the profiler results viewer"""
        return

    def getShortName(self):
        """Tells the display name"""
        return "Profiling results"

    def setShortName(self, name):
        """Sets the display name - not applicable"""
        raise Exception("Setting a file name for profiler "
                        "results is not applicable")



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The original code is taken from qutepart project and adopted for Codimension
# See https://github.com/andreikop/qutepart
#


"""Line numbers margin"""

from ui.qt import QWidget, QPainter, Qt
from qutepart.margins import MarginBase
from utils.misc import extendInstance
from utils.globals import GlobalData
from utils.colorfont import getZoomedMarginFont


class CDMLineNumberMargin(QWidget):

    """Line number area widget"""

    _LEFT_MARGIN = 5
    _RIGHT_MARGIN = 3

    def __init__(self, parent):
        QWidget.__init__(self, parent)

        extendInstance(self, MarginBase)
        MarginBase.__init__(self, parent, 'cdm_line_number_margin', 0)

        self.__bgColor = GlobalData().skin['marginPaper']
        self.__fgColor = GlobalData().skin['marginColor']

        self.__width = self.__calculateWidth()
        self.onTextZoomChanged()

        # The width needs to be re-calculated when the margin is drawn the
        # first time. The problem is that if the widget is not on the screen
        # then the font metrics are not calculated properly and thus the width
        # is not shown right. What I observed is an offset up to 2 pixels.
        self.__firstTime = True

        self._qpart.blockCountChanged.connect(self.__updateWidth)

    # Arguments: newBlockCount
    def __updateWidth(self, _=None):
        """Updates the margin width"""
        newWidth = self.__calculateWidth()
        if newWidth != self.__width:
            self.__width = newWidth
            self._qpart.updateViewport()

    def paintEvent(self, event):
        """Paints the margin"""
        if self.__firstTime:
            self.__updateWidth()
            self.__firstTime = False

        painter = QPainter(self)
        painter.fillRect(event.rect(), self.__bgColor)
        painter.setPen(self.__fgColor)

        block = self._qpart.firstVisibleBlock()
        blockNumber = block.blockNumber()
        top = int(self._qpart.blockBoundingGeometry(block).
                  translated(self._qpart.contentOffset()).top())
        bottom = top + int(self._qpart.blockBoundingRect(block).height())

        boundingRect = self._qpart.blockBoundingRect(block)
        availableWidth = self.__width - self._RIGHT_MARGIN - self._LEFT_MARGIN

        # The margin font could be smaller than the main area font
        topShift = int((self._qpart.fontMetrics().height() -
                        self.fontMetrics().height()) / 2)
        if topShift < 0:
            topShift = 0

        availableHeight = self._qpart.fontMetrics().height()
        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                number = str(blockNumber + 1)
                painter.drawText(self._LEFT_MARGIN, top + topShift,
                                 availableWidth,
                                 availableHeight,
                                 Qt.AlignRight, number)
            block = block.next()
            boundingRect = self._qpart.blockBoundingRect(block)
            top = bottom
            bottom = top + int(boundingRect.height())
            blockNumber += 1

    def __calculateWidth(self):
        """Calculates the margin width"""
        digits = len(str(max(1, self._qpart.blockCount())))
        digitsWidth = self.fontMetrics().width('9') * digits
        return self._LEFT_MARGIN + digitsWidth + self._RIGHT_MARGIN

    def width(self):
        """Desired width. Includes text and margins"""
        return self.__width

    def setFont(self, font):
        """Overloaded to adjust the width if needed"""
        QWidget.setFont(self, font)
        self.__updateWidth()

    def setBackgroundColor(self, color):
        """Sets the new background color"""
        if self.__bgColor != color:
            self.__bgColor = color
            self.update()

    def setForegroundColor(self, color):
        """Sets the new foreground color"""
        if self.__fgColor != color:
            self.__fgColor = color
            self.update()

    def onTextZoomChanged(self):
        """Triggered when a zoom has been changed"""
        self.setFont(getZoomedMarginFont())

    def onClose(self):
        """The editor is going to be closed"""
        self._qpart.blockCountChanged.disconnect(self.__updateWidth)

        # Release the editor reference
        self._qpart = None



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2018  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Markdown viewer tab widget"""

import logging
import os.path
from editor.mdwidget import MDViewer
from utils.globals import GlobalData
from utils.fileutils import getFileContent, getFileProperties, isMarkdownMime
from utils.md import renderMarkdown
from utils.pixmapcache import getIcon
from .qt import (QWidget, Qt, pyqtSignal, QAction, QSize, QToolBar,
                 QPrintDialog, QDialog, QHBoxLayout)
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase
from .viewhistory import ViewEntry, ViewHistory


class MDFullViewer(MDViewer):

    """Markdown viewer specific for non-editing mode (full view)"""

    def __init__(self, history, parent):
        MDViewer.__init__(self, parent)
        self.__history = history
        self.__parentWidget = parent

    def _onAnchorClicked(self, link):
        """Overwritten URL click handler"""
        if link == 'javascript:history.back()' or link == 'history.back()':
            if self.__parentWidget.viewerHistory.backAvailable():
                self.__parentWidget.onBack()
            else:
                logging.warning('No step back avaialable')
            return

        if link == 'javascript:history.forward()' or link == 'history.forward()':
            if self.__parentWidget.viewerHistory.forwardAvailable():
                self.__parentWidget.onForward()
            else:
                logging.warning('No step forward available')
            return

        fileName, anchorOrLine = self._resolveLink(link)
        if fileName:
            mime, _, _ = getFileProperties(fileName)
            if isMarkdownMime(mime):
                self.__parentWidget.updateCurrentHistoryPosition()
                self.__parentWidget.setFileName(fileName)
            else:
                # It is a resolved link to some kind of non-markdown file
                GlobalData().mainWindow.openFile(fileName, anchorOrLine)


class MarkdownTabWidget(QWidget, MainWindowTabWidgetBase):

    """Markdown widget"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, readOnly, parent=None):
        MainWindowTabWidgetBase.__init__(self)
        QWidget.__init__(self, parent)

        self.viewerHistory = ViewHistory()
        self.viewerHistory.historyChanged.connect(self.__onHistoryChanged)
        self.__viewer = MDFullViewer(self.viewerHistory, self)
        self.__viewer.sigEscapePressed.connect(self.__onEsc)
        self.__readOnly = readOnly
        self.__fName = None

        self.__createLayout()

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def __createLayout(self):
        """Creates the toolbar and layout"""
        printButton = QAction(getIcon('printer.png'), 'Print', self)
        printButton.triggered.connect(self.__onPrint)

        fixedSpacer = QWidget()
        fixedSpacer.setFixedHeight(16)

        self.__backButton = QAction(getIcon('mdback.png'), 'Back', self)
        self.__backButton.triggered.connect(self.onBack)
        self.__backButton.setEnabled(False)

        self.__fwdButton = QAction(getIcon('mdfwd.png'), 'Forward', self)
        self.__fwdButton.triggered.connect(self.onForward)
        self.__fwdButton.setEnabled(False)

        if not self.__readOnly:
            self.__switchToEditButton = QAction(getIcon('switchtoedit.png'),
                                                'Switch to Editing', self)
            self.__switchToEditButton.triggered.connect(self.__onSwitchToEdit)
            self.__switchToEditButton.setEnabled(True)

            self.__reloadButton = QAction(getIcon('mdreload.png'),
                                          'Reload content', self)
            self.__reloadButton.triggered.connect(self.__onReload)
            self.__reloadButton.setEnabled(True)

        # Toolbar
        toolbar = QToolBar(self)
        toolbar.setOrientation(Qt.Vertical)
        toolbar.setMovable(False)
        toolbar.setAllowedAreas(Qt.RightToolBarArea)
        toolbar.setIconSize(QSize(16, 16))
        toolbar.setFixedWidth(28)
        toolbar.setContentsMargins(0, 0, 0, 0)

        toolbar.addAction(printButton)
        toolbar.addWidget(fixedSpacer)
        toolbar.addAction(self.__backButton)
        toolbar.addAction(self.__fwdButton)

        if not self.__readOnly:
            fixedSpacer2 = QWidget()
            fixedSpacer2.setFixedHeight(16)
            toolbar.addWidget(fixedSpacer2)
            toolbar.addAction(self.__switchToEditButton)

            fixedSpacer3 = QWidget()
            fixedSpacer3.setFixedHeight(16)
            toolbar.addWidget(fixedSpacer3)
            toolbar.addAction(self.__reloadButton)

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)
        hLayout.addWidget(self.__viewer)
        hLayout.addWidget(toolbar)

        self.setLayout(hLayout)

    def setFocus(self):
        """Overridden setFocus"""
        self.__viewer.setFocus()

    def isReadOnly(self):
        """True if it is a read only mode"""
        return self.__readOnly

    def __onPrint(self):
        """Print the markdown page"""
        dialog = QPrintDialog(self)
        if dialog.exec_() == QDialog.Accepted:
            printer = dialog.printer()
            self.__viewer.print_(printer)

    def updateCurrentHistoryPosition(self):
        """Updates the current history entry position value"""
        pos = tuple(self.__viewer.getScrollbarPositions())
        self.viewerHistory.getCurrentEntry().pos = pos

    def onBack(self):
        """Back in the history"""
        if self.viewerHistory.backAvailable():
            self.updateCurrentHistoryPosition()
            self.viewerHistory.stepBack()
            entry = self.viewerHistory.getCurrentEntry()
            self.setFileName(entry.fName, True)
            self.__viewer.setScrollbarPositions(entry.pos[0], entry.pos[1])

    def onForward(self):
        """Forward in the history"""
        if self.viewerHistory.forwardAvailable():
            self.updateCurrentHistoryPosition()
            self.viewerHistory.stepForward()
            entry = self.viewerHistory.getCurrentEntry()
            self.setFileName(entry.fName, True)
            self.__viewer.setScrollbarPositions(entry.pos[0], entry.pos[1])

    def __onSwitchToEdit(self):
        """Switch to editing mode"""
        fName = self.getFileName()
        if fName:
            mainWindow = GlobalData().mainWindow
            mainWindow.em.onCloseTab()
            mainWindow.em.openFile(fName, -1)

    def __onReload(self):
        """Reloads the file content"""
        if self.getFileName():
            hPos, vPos = self.__viewer.getScrollbarPositions()
            self.setFileName(self.getFileName(), True)
            self.__viewer.setScrollbarPositions(hPos, vPos)

    def __onHistoryChanged(self):
        """Updates the back and forward buttons"""
        self.__backButton.setEnabled(self.viewerHistory.backAvailable())
        self.__fwdButton.setEnabled(self.viewerHistory.forwardAvailable())

    # Mandatory interface part is below

    def isModified(self):
        """Tells if the file is modified"""
        return False

    def getRWMode(self):
        """Tells if the file is read only"""
        return "RO"

    def getType(self):
        """Tells the widget type"""
        return MainWindowTabWidgetBase.MDViewer

    def getLanguage(self):
        """Tells the content language"""
        return "Markdown"

    def setFileName(self, path, fromHistory=False):
        """Sets the file name"""
        try:
            content = getFileContent(path)
        except Exception as exc:
            logging.error(str(exc))
            return

        renderedText, errors, warnings = renderMarkdown(self.getUUID(),
                                                        content, path)
        if errors:
            for error in errors:
                logging.error(error)
            return

        if not fromHistory:
            # Need to add an entry for the new item
            newEntry = ViewEntry(path, (0, 0))
            self.viewerHistory.addEntry(newEntry)

        self.__fName = path
        self.__viewer.setHtml(renderedText)
        if warnings:
            for warning in warnings:
                logging.warning(warning)

        mainWindow = GlobalData().mainWindow
        mainWindow.em.setTabText(mainWindow.em.currentIndex(),
                                 self.getShortName())
        mainWindow.em.updateStatusBar()

    def getFileName(self):
        """Tells what file name of the widget"""
        return self.__fName

    def setEncoding(self, newEncoding):
        """Sets the new encoding - not applicable for the markdown viewer"""
        return

    def getShortName(self):
        """Tells the display name"""
        if self.__fName:
            return os.path.basename(self.__fName)
        return 'n/a'

    def setShortName(self, name):
        """Sets the display name - not applicable"""
        raise Exception("Setting a short name for a markdown viewer "
                        "is not applicable")



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Base and auxiliary classes for G/F/C browsers"""

import os
from utils.globals import GlobalData
from utils.project import CodimensionProject
from utils.fileutils import isPythonFile
from .qt import (Qt, QModelIndex, QRegExp, pyqtSignal, QSortFilterProxyModel,
                 QAbstractItemView, QApplication, QTreeView)
from .itemdelegates import NoOutlineHeightDelegate
from .viewitems import (DirectoryItemType, SysPathItemType, GlobalsItemType,
                        ImportsItemType, FunctionsItemType, ClassesItemType,
                        StaticAttributesItemType, InstanceAttributesItemType,
                        FunctionItemType, ClassItemType)


class ObjectsBrowserSortFilterProxyModel(QSortFilterProxyModel):

    """Objects (globals, functions, classes) browser sort filter proxy model
       implementation. It allows filtering basing on top level items.
    """

    def __init__(self, parent=None):
        QSortFilterProxyModel.__init__(self, parent)
        self.__sortColumn = None    # Avoid pylint complains
        self.__sortOrder = None     # Avoid pylint complains

        self.__filters = []
        self.__filtersCount = 0
        self.__sourceModelRoot = None

    def sort(self, column, order):
        """Sorts the items"""
        self.__sortColumn = column
        self.__sortOrder = order
        QSortFilterProxyModel.sort(self, column, order)

    def lessThan(self, left, right):
        """Sorts the displayed items"""
        lmodel = left.model()
        lhs = lmodel and lmodel.item(left) or None
        if lhs:
            rmodel = right.model()
            rhs = rmodel and rmodel.item(right) or None
            if rhs:
                return lhs.lessThan(rhs, self.__sortColumn, self.__sortOrder)
        return False

    def item(self, index):
        """Provides a reference to the item"""
        if not index.isValid():
            return None

        sourceIndex = self.mapToSource(index)
        return self.sourceModel().item(sourceIndex)

    def hasChildren(self, parent=QModelIndex()):
        """Checks the presence of the child items"""
        sourceIndex = self.mapToSource(parent)
        return self.sourceModel().hasChildren(sourceIndex)

    def setFilter(self, text):
        """Sets the new filters"""
        self.__filters = []
        self.__filtersCount = 0
        self.__sourceModelRoot = None
        for part in str(text).strip().split():
            regexp = QRegExp(part, Qt.CaseInsensitive, QRegExp.RegExp2)
            self.__filters.append(regexp)
            self.__filtersCount += 1
        self.__sourceModelRoot = self.sourceModel().rootItem

    def filterAcceptsRow(self, sourceRow, sourceParent):
        """Filters rows"""
        if self.__filtersCount == 0 or self.__sourceModelRoot is None:
            return True     # No filters

        sindex = self.sourceModel().index(sourceRow, 0, sourceParent)
        if not sindex.isValid():
            return False

        sitem = self.sourceModel().item(sindex)
        if not sitem is None:
            parent = sitem.parent()
            if not parent is None:
                if parent.parentItem is None:
                    # Filter top level items only
                    nameToMatch = sitem.sourceObj.name
                    for regexp in self.__filters:
                        if regexp.indexIn(nameToMatch) == -1:
                            return False

        # Show all the nested items
        return True


class ObjectsBrowser(QTreeView):

    """Common functionality of the G/F/C browsers"""

    sigOpeningItem = pyqtSignal(str, int)
    sigSelectionChanged = pyqtSignal(QModelIndex)
    sigModelFilesChanged = pyqtSignal()

    def __init__(self, sourceModel, parent=None):
        QTreeView.__init__(self, parent)

        self.__model = sourceModel
        self.__sortModel = ObjectsBrowserSortFilterProxyModel()
        self.__sortModel.setDynamicSortFilter(True)
        self.__sortModel.setSourceModel(self.__model)
        self.setModel(self.__sortModel)
        self.contextItem = None

        self.activated.connect(self._openItem)
        self.expanded.connect(self._resizeColumns)
        self.collapsed.connect(self._resizeColumns)

        GlobalData().project.sigFSChanged.connect(self.onFSChanged)
        self.__model.modelReset.connect(self.updateCounter)

        self.setRootIsDecorated(True)
        self.setAlternatingRowColors(True)
        self.setUniformRowHeights(True)
        self.setItemDelegate(NoOutlineHeightDelegate(4))

        header = self.header()
        header.setSortIndicator(0, Qt.AscendingOrder)
        header.setSortIndicatorShown(True)
        header.setSectionsClickable(True)

        self.setSortingEnabled(True)

        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.layoutDisplay()

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)

    def updateCounter(self, parent=None, start=0, end=0):
        """Updates the header with the currently visible and total items"""
        del parent  # unused argument
        del start   # unused argument
        del end     # unused argument
        text = "Name (" + str(self.model().rowCount()) + " of " + \
                          str(self.__model.totalRowCount()) + ")"
        self.__model.updateRootData(0, text)

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.__model.beginResetModel()
            self.__model.endResetModel()
            self.layoutDisplay()
            self.updateCounter()

    def setFilter(self, text):
        """Sets the new filter for items"""
        # Notify the filtering model of the new filters
        self.model().setFilter(text)

        # This is to trigger filtering - ugly but I don't know how else
        self.model().setFilterRegExp('')

        # No need to resort but need to resize columns
        self._resizeColumns(QModelIndex())

    def layoutDisplay(self):
        """Performs the layout operation"""
        self._resizeColumns(QModelIndex())
        self._resort()

    def _resizeColumns(self, index):
        """Resizes the view when items get expanded or collapsed"""
        rowCount = self.model().rowCount()
        columnCount = self.model().columnCount()
        self.header().setStretchLastSection(rowCount == 0)

        index = 0
        while index < columnCount - 1:
            width = max(100, self.sizeHintForColumn(index))
            self.header().resizeSection(index, width)
            index += 1

        # The last column is 'Line' so it should be narrower
        width = max(40, self.sizeHintForColumn(index))
        self.header().resizeSection(index, width)

    def _resort(self):
        """Re-sorts the tree"""
        self.model().sort(self.header().sortIndicatorSection(),
                          self.header().sortIndicatorOrder())

    def mouseDoubleClickEvent(self, mouseEvent):
        """Reimplemented to disable expanding/collapsing of items on dbl click.

        Instead the double-clicked entry is opened.
        """
        index = self.indexAt(mouseEvent.pos())
        if not index.isValid():
            return

        item = self.model().item(index)
        if item.itemType in [GlobalsItemType,
                             ImportsItemType, FunctionsItemType,
                             ClassesItemType, StaticAttributesItemType,
                             InstanceAttributesItemType,
                             DirectoryItemType, SysPathItemType]:
            # This will return the first column index regardless in what
            # column the double click happened
            index = self.selectedIndexes()[0]
            if self.isExpanded(index):
                self.collapse(index)
            else:
                self.expand(index)
        else:
            self.openItem(item)

    def _openItem(self):
        """Triggers when an item is clicked or double clicked"""
        item = self.model().item(self.selectedIndexes()[0])
        self.openItem(item)

    def openItem(self, item):
        """Handles the case when an item is activated"""
        if item.itemType in [GlobalsItemType,
                             ImportsItemType, FunctionsItemType,
                             ClassesItemType, StaticAttributesItemType,
                             InstanceAttributesItemType,
                             DirectoryItemType, SysPathItemType]:
            return
        path = item.getPath()
        line = item.data(2)
        self.sigOpeningItem.emit(str(path), line)
        GlobalData().mainWindow.openFile(path, line)

    @staticmethod
    def getDisassembled(item):
        """Handles showing disassembled code"""
        if item.itemType not in [FunctionItemType, ClassItemType]:
            return

        path = item.getPath()
        qualifiedName = item.getQualifiedName()
        GlobalData().mainWindow.showDisassembler(path, qualifiedName)

    def copyToClipboard(self):
        """Copies the path to the file where the element is to the clipboard"""
        item = self.model().item(self.currentIndex())
        QApplication.clipboard().setText(item.getPath())

    def onFileUpdated(self, fileName):
        """Triggered when the file is updated"""
        if not GlobalData().project.isProjectFile(fileName):
            # Not a project file
            return

        if not isPythonFile(fileName):
            return

        if self.__model.onFileUpdated(fileName):
            # Need resort and counter updates
            self.layoutDisplay()
            self.updateCounter()
            self.model().setFilterRegExp("")

    def onFSChanged(self, items):
        """Triggered when filesystem has changes"""
        addedPythonFiles = []
        deletedPythonFiles = []

        for path in items:
            path = str(path)
            if path.endswith(os.path.sep):
                continue    # dirs are out of interest
            if path.startswith('+'):
                path = path[1:]
                if not isPythonFile(path):
                    continue
                addedPythonFiles.append(path)
            else:
                path = path[1:]
                if not isPythonFile(path):
                    continue
                deletedPythonFiles.append(path)
        if addedPythonFiles or deletedPythonFiles:
            if self.__model.onFSChanged(addedPythonFiles,
                                        deletedPythonFiles):
                # Need resort and counter updates
                self.layoutDisplay()
                self.updateCounter()
                self.model().setFilterRegExp("")
                self.sigModelFilesChanged.emit()

    def selectionChanged(self, selected, deselected):
        """Slot is called when the selection has been changed"""
        if selected.indexes():
            # The objects browsers may have no more than one selected item
            self.sigSelectionChanged.emit(selected.indexes()[0])
        else:
            self.sigSelectionChanged.emit(QModelIndex())
        QTreeView.selectionChanged(self, selected, deselected)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Provides the storage for the user assigned file encoding"""

import os.path
from .fileutils import loadJSON, saveJSON


class FileEncodings:

    """Loads/stores/saves the file encodingss"""

    def __init__(self):
        # file name -> encoding
        self.__encodings = {}
        self.__encFileName = None

    def reset(self):
        """Un-binds from the file system"""
        self.__encodings = {}
        self.__encFileName = None

    def setup(self, dirName):
        """Binds the parameters to a disk file"""
        # Just in case - flush the previous data if they were bound
        FileEncodings.save(self)

        dirName = os.path.realpath(dirName)
        if not dirName.endswith(os.path.sep):
            dirName += os.path.sep
        if not os.path.isdir(dirName):
            raise Exception('Directory name is expected for file '
                            'encodings. The given ' + dirName + ' is not.')

        self.__encFileName = dirName + 'encodings.json'
        if os.path.exists(self.__encFileName):
            FileEncodings.load(self)

    def load(self):
        """Loads the saved encodings file"""
        if self.__encFileName:
            self.__encodings = loadJSON(self.__encFileName,
                                        'file encodings', {})

    def save(self):
        """Saves the encodings into a file"""
        if self.__encFileName:
            saveJSON(self.__encFileName, self.__encodings, 'file encodings')

    def getFileEncoding(self, fileName):
        """Provides None if not found"""
        return self.__encodings.get(fileName, None)

    def setFileEncoding(self, fileName, encoding):
        """Sets the encoding for the file"""
        if encoding:
            self.__encodings[fileName] = encoding
            FileEncodings.save(self)
        else:
            if self.__encodings.pop(fileName, None):
                FileEncodings.save(self)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""find in files dialog"""


from os import listdir
from os.path import sep, isabs, exists, isdir, normpath, isfile, realpath
import re
import time
import logging
from html import escape
from utils.globals import GlobalData
from utils.settings import Settings
from utils.fileutils import (isFileSearchable, isPythonFile, getFileContent,
                             resolveLink)
from utils.diskvaluesrelay import getFindInFilesHistory, setFindInFilesHistory
from cdmpyparser import getBriefModuleInfoFromMemory
from .qt import (QCursor, Qt, QDialog, QDialogButtonBox, QVBoxLayout,
                 QSizePolicy, QLabel, QProgressBar, QApplication, QComboBox,
                 QGridLayout, QHBoxLayout, QCheckBox, QRadioButton,
                 QGroupBox, QPushButton, QFileDialog)
from .labels import FitPathLabel
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase


class Match:

    """Stores info about one match in a file"""

    def __init__(self, line, start, finish):
        self.line = line        # Matched line
        self.start = start      # Match start pos
        self.finish = finish    # Match end pos
        self.tooltip = "not implemented"
        self.text = ""


def getSearchItemIndex(items, fileName):
    """Provides the search item index basing on the file name"""
    index = 0
    for item in items:
        if item.fileName == fileName:
            return index
        index += 1
    return -1


class ItemToSearchIn:

    """Stores information about one item to search in"""

    contextLines = 15

    def __init__(self, fname, bufferID):
        self.fileName = fname       # Could be absolute -> for existing files
                                    # or relative -> for newly created
        self.bufferUUID = bufferID  # Non empty for currently opened files
        self.tooltip = ""           # For python files only -> docstring
        self.matches = []

    def addMatch(self, name, lineNumber, customMessage=None):
        """Used to add a match which was found outside of find in files"""
        match = Match(lineNumber, 0, 0)

        # Load the file and identify matched line and tooltip
        try:
            if self.bufferUUID != "":
                mainWindow = GlobalData().mainWindow
                widget = mainWindow.getWidgetByUUID(self.bufferUUID)
                if widget is not None:
                    content = widget.getEditor().lines
                else:
                    raise Exception('Inconsistency. Buffer disappeared.')
            else:
                content = getFileContent(self.fileName).splitlines()
            self.__fillInMatch(match, content, name, lineNumber, customMessage)
        except Exception as exc:
            logging.error('Error adding match: %s', str(exc))
        self.matches.append(match)

    def __fillInMatch(self, match, content, name, lineNumber,
                      customMessage=None):
        """Fills in the match fields from the content"""
        # Form the regexp corresponding to a single word search
        line = content[lineNumber - 1]
        if customMessage:
            match.text = customMessage
        else:
            match.text = line.strip()

        if name:
            regexpText = re.escape(name)
            regexpText = "\\b%s\\b" % regexpText
            flags = re.UNICODE
            searchRegexp = re.compile(regexpText, flags)

            contains = searchRegexp.search(line)
            match.start = contains.start()
            match.finish = contains.end()
        else:
            match.start = 0
            match.finish = len(line)

        match.tooltip = self.__buildTooltip(content, lineNumber - 1,
                                            len(content),
                                            match.start, match.finish)

        self.__extractDocstring(content)

    def search(self, expression):
        """Perform search within this item"""
        self.matches = []
        if self.bufferUUID != "":
            # Search item is the currently loaded buffer
            mainWindow = GlobalData().mainWindow
            widget = mainWindow.getWidgetByUUID(self.bufferUUID)
            if widget is not None:
                # Search in the buffer

                self.__lookThroughLines(widget.getEditor().lines, expression)
                return

        # Here: there were no buffer or have not found it
        #       try searching in a file
        if not isabs(self.fileName) or not exists(self.fileName):
            # Unfortunately not all network file systems report the
            # fact that a file has been deleted from the disk so
            # let's simply ignore such files
            return

        # File exists, search in the file
        try:
            content = getFileContent(self.fileName).splitlines()
            self.__lookThroughLines(content, expression)
        except Exception as exc:
            logging.error('Error searching in %s: %s', self.fileName, str(exc))

    def __buildTooltip(self, content, lineIndex, totalLines,
                       startPos, finishPos):
        """Forms the tooltip for the given match"""
        start, end = self.__calculateContextStart(lineIndex, totalLines)
        lines = content[start:end]
        matchIndex = lineIndex - start

        # Avoid incorrect tooltips for HTML/XML files
        for index in range(0, len(lines)):
            if index != matchIndex:
                lines[index] = escape(lines[index])

        lines[matchIndex] = \
            escape(lines[matchIndex][:startPos]) + \
            "<b>" + \
            escape(lines[matchIndex][startPos:finishPos]) + \
            "</b>" + \
            escape(lines[matchIndex][finishPos:])

        # Strip empty lines at the end and at the beginning
        index = len(lines) - 1
        while index >= 0:
            if lines[index].strip() == '':
                del lines[index]
                index -= 1
                continue
            break
        while len(lines) > 0:
            if lines[0].strip() == '':
                del lines[0]
                continue
            break

        return '<p>' + '<br/>'.join(lines).replace(' ', '&nbsp;') + '</p>'

    def __lookThroughLines(self, content, expression):
        """Searches through all the given lines"""
        lineIndex = 0
        totalLines = len(content)
        while lineIndex < totalLines:
            line = content[lineIndex]
            contains = expression.search(line)
            if contains:
                match = Match(lineIndex + 1, contains.start(), contains.end())
                match.text = line.strip()
                match.tooltip = self.__buildTooltip(content, lineIndex,
                                                    totalLines,
                                                    match.start, match.finish)
                self.matches.append(match)
                if len(self.matches) > 1024:
                    # Too much entries, stop here
                    logging.warning("More than 1024 matches in %s. Stop "
                                    "further search in this file.",
                                    self.fileName)
                    break
            lineIndex += 1

        # Extract docsting if applicable
        if len(self.matches) > 0:
            self.__extractDocstring(content)

    def __extractDocstring(self, content):
        """Extracts a docstring and sets it as a tooltip if needed"""
        if self.tooltip != "":
            return

        if isPythonFile(self.fileName):
            info = getBriefModuleInfoFromMemory("\n".join(content))
            self.tooltip = ""
            if info.docstring is not None:
                self.tooltip = info.docstring.text

    @staticmethod
    def __calculateContextStart(matchedLine, totalLines):
        """Calculates the start line number for the context tooltip"""
        # matchedLine is a zero based index
        if ItemToSearchIn.contextLines >= totalLines:
            return 0, totalLines

        start = matchedLine - int(ItemToSearchIn.contextLines / 2)
        if start < 0:
            start = 0
        end = start + ItemToSearchIn.contextLines
        if end < totalLines:
            return start, end
        return totalLines - ItemToSearchIn.contextLines, totalLines


class FindInFilesDialog(QDialog):

    """find in files dialog implementation"""

    IN_PROJECT = 0
    IN_DIRECTORY = 1
    IN_OPEN_FILES = 2

    def __init__(self, where, what=None, dirPath=None):
        QDialog.__init__(self, GlobalData().mainWindow)

        mainWindow = GlobalData().mainWindow
        self.editorsManager = mainWindow.editorsManagerWidget.editorsManager

        self.__cancelRequest = False
        self.__inProgress = False
        self.searchRegexp = None
        self.searchResults = []

        # Avoid pylint complains
        self.findCombo = None
        self.caseCheckBox = None
        self.wordCheckBox = None
        self.regexpCheckBox = None
        self.projectRButton = None
        self.openFilesRButton = None
        self.dirRButton = None
        self.dirEditCombo = None
        self.dirSelectButton = None
        self.filterCombo = None
        self.fileLabel = None
        self.progressBar = None
        self.findButton = None

        self.__createLayout()
        self.setWindowTitle("Find in files")

        self.__maxEntries = Settings()['maxSearchEntries']

        # Restore the combo box values
        # [ {'term': ., 'dir': ., 'filters': .,
        #    'cbCase': ., 'cbWord': ., 'cbRegexp': .,
        #    'rbProject': ., 'rbOpen': ., 'rbDir': .}, ... ]
        self.__history = getFindInFilesHistory()
        self.__populateHistory()
        self.findCombo.setEditText('')
        self.dirEditCombo.setEditText('')
        self.filterCombo.setEditText('')

        if where == self.IN_PROJECT:
            self.setSearchInProject(what)
        elif where == self.IN_DIRECTORY:
            self.setSearchInDirectory(what, dirPath)
        else:
            self.setSearchInOpenFiles(what)

    def __serialize(self):
        """Serializes the current search parameters"""
        termText = self.findCombo.currentText()
        filtText = self.__normalizeFilters(self.filterCombo.currentText())
        dirText = ''
        if self.dirRButton.isChecked():
            dirText = self.dirEditCombo.currentText().strip()

        return {'term': termText,
                'dir': dirText,
                'filters': filtText,
                'cbCase': self.caseCheckBox.isChecked(),
                'cbWord': self.wordCheckBox.isChecked(),
                'cbRegexp': self.regexpCheckBox.isChecked(),
                'rbProject': self.projectRButton.isChecked(),
                'rbOpen': self.openFilesRButton.isChecked(),
                'rbDir': self.dirRButton.isChecked()}

    def __deserialize(self, item):
        """Deserializes the history item"""
        self.findCombo.setEditText(item['term'])
        self.dirEditCombo.setEditText(item['dir'])
        self.filterCombo.setEditText(item['filters'])
        self.caseCheckBox.setChecked(item['cbCase'])
        self.wordCheckBox.setChecked(item['cbWord'])
        self.regexpCheckBox.setChecked(item['cbRegexp'])

        self.projectRButton.setChecked(item['rbProject'])
        self.openFilesRButton.setChecked(item['rbOpen'])
        self.dirRButton.setChecked(item['rbDir'])

        self.dirEditCombo.setEnabled(item['rbDir'])
        self.dirSelectButton.setEnabled(item['rbDir'])

    def __populateHistory(self):
        """Populates the search history in the combo boxes"""
        # No need to react to the change of the current index
        self.findCombo.currentIndexChanged[int].disconnect(
            self.__whatIndexChanged)
        index = 0
        for props in self.__history:
            self.findCombo.addItem(props['term'], index)
            directory = props['dir']
            if directory:
                self.dirEditCombo.addItem(directory)
            filt = props['filters']
            if filt:
                self.filterCombo.addItem(filt)
            index += 1
        # Restore the handler
        self.findCombo.currentIndexChanged[int].connect(
            self.__whatIndexChanged)

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(600, 300)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)
        gridLayout = QGridLayout()

        # Combo box for the text to search
        findLabel = QLabel(self)
        findLabel.setText("Find text:")
        self.findCombo = QComboBox(self)
        self.__tuneCombo(self.findCombo)
        self.findCombo.lineEdit().setToolTip(
            "Regular expression to search for")
        self.findCombo.editTextChanged.connect(self.__someTextChanged)
        self.findCombo.currentIndexChanged[int].connect(
            self.__whatIndexChanged)

        gridLayout.addWidget(findLabel, 0, 0, 1, 1)
        gridLayout.addWidget(self.findCombo, 0, 1, 1, 1)
        verticalLayout.addLayout(gridLayout)

        # Check boxes
        horizontalCBLayout = QHBoxLayout()
        self.caseCheckBox = QCheckBox(self)
        self.caseCheckBox.setText("Match &case")
        horizontalCBLayout.addWidget(self.caseCheckBox)
        self.wordCheckBox = QCheckBox(self)
        self.wordCheckBox.setText("Match whole &word")
        horizontalCBLayout.addWidget(self.wordCheckBox)
        self.regexpCheckBox = QCheckBox(self)
        self.regexpCheckBox.setText("Regular &expression")
        horizontalCBLayout.addWidget(self.regexpCheckBox)

        verticalLayout.addLayout(horizontalCBLayout)

        # Files groupbox
        filesGroupbox = QGroupBox(self)
        filesGroupbox.setTitle("Find in")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            filesGroupbox.sizePolicy().hasHeightForWidth())
        filesGroupbox.setSizePolicy(sizePolicy)

        gridLayoutFG = QGridLayout(filesGroupbox)
        self.projectRButton = QRadioButton(filesGroupbox)
        self.projectRButton.setText("&Project")
        gridLayoutFG.addWidget(self.projectRButton, 0, 0)
        self.projectRButton.clicked.connect(self.__projectClicked)

        self.openFilesRButton = QRadioButton(filesGroupbox)
        self.openFilesRButton.setText("&Opened files only")
        gridLayoutFG.addWidget(self.openFilesRButton, 1, 0)
        self.openFilesRButton.clicked.connect(self.__openFilesOnlyClicked)

        self.dirRButton = QRadioButton(filesGroupbox)
        self.dirRButton.setText("&Directory tree")
        gridLayoutFG.addWidget(self.dirRButton, 2, 0)
        self.dirRButton.clicked.connect(self.__dirClicked)

        self.dirEditCombo = QComboBox(filesGroupbox)
        self.__tuneCombo(self.dirEditCombo)
        self.dirEditCombo.lineEdit().setToolTip("Directory to search in")
        gridLayoutFG.addWidget(self.dirEditCombo, 2, 1)
        self.dirEditCombo.editTextChanged.connect(self.__someTextChanged)

        self.dirSelectButton = QPushButton(filesGroupbox)
        self.dirSelectButton.setText("...")
        gridLayoutFG.addWidget(self.dirSelectButton, 2, 2)
        self.dirSelectButton.clicked.connect(self.__selectDirClicked)

        filterLabel = QLabel(filesGroupbox)
        filterLabel.setText("Files filter:")
        gridLayoutFG.addWidget(filterLabel, 3, 0)
        self.filterCombo = QComboBox(filesGroupbox)
        self.__tuneCombo(self.filterCombo)
        self.filterCombo.lineEdit().setToolTip("File names regular expression")
        gridLayoutFG.addWidget(self.filterCombo, 3, 1)
        self.filterCombo.editTextChanged.connect(self.__someTextChanged)

        verticalLayout.addWidget(filesGroupbox)

        # File label
        self.fileLabel = FitPathLabel(parent=self)
        self.fileLabel.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed)
        verticalLayout.addWidget(self.fileLabel)

        # Progress bar
        self.progressBar = QProgressBar(self)
        self.progressBar.setValue(0)
        self.progressBar.setOrientation(Qt.Horizontal)
        verticalLayout.addWidget(self.progressBar)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Cancel)
        self.findButton = buttonBox.addButton("Find",
                                              QDialogButtonBox.AcceptRole)
        self.findButton.setDefault(True)
        self.findButton.clicked.connect(self.__process)
        verticalLayout.addWidget(buttonBox)

        buttonBox.rejected.connect(self.__onClose)

    @staticmethod
    def __tuneCombo(comboBox):
        """Sets the common settings for a combo box"""
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            comboBox.sizePolicy().hasHeightForWidth())
        comboBox.setSizePolicy(sizePolicy)
        comboBox.setEditable(True)
        comboBox.setInsertPolicy(QComboBox.InsertAtTop)
        comboBox.setCompleter(None)
        comboBox.setDuplicatesEnabled(False)

    def __onClose(self):
        """Triggered when the close button is clicked"""
        self.__cancelRequest = True
        if not self.__inProgress:
            self.close()

    def __historyIndexByWhat(self, what):
        """Provides the history index by 'what' value"""
        if what:
            for index in range(self.findCombo.count()):
                if self.findCombo.itemText(index) == what:
                    return index, self.findCombo.itemData(index)
        return None, None

    @staticmethod
    def __indexByText(combo, text):
        """Provides the text entry index"""
        index = 0
        for index in range(combo.count()):
            if combo.itemText(index) == text:
                return index
            index += 1
        return -1

    def setSearchInProject(self, what=None):
        """Set search ready for the whole project"""
        if not GlobalData().project.isLoaded():
            # No project loaded, fallback to opened files
            self.setSearchInOpenFiles(what)
            return

        # Select the project radio button
        self.projectRButton.setEnabled(True)
        self.projectRButton.setChecked(True)
        self.dirEditCombo.setEnabled(False)
        self.dirSelectButton.setEnabled(False)

        openedFiles = self.editorsManager.getTextEditors()
        self.openFilesRButton.setEnabled(len(openedFiles) != 0)

        if what:
            # Pick up the history values if so
            comboIndex, historyIndex = self.__historyIndexByWhat(what)
            if historyIndex is not None:
                self.__deserialize(self.__history[historyIndex])
                self.findCombo.setCurrentIndex(comboIndex)
            else:
                self.findCombo.setCurrentText(what)
            self.findCombo.lineEdit().selectAll()
        self.findCombo.setFocus()

        # Check searchability
        self.__testSearchability()

    def setSearchInOpenFiles(self, what=None):
        """Sets search ready for the opened files"""
        openedFiles = self.editorsManager.getTextEditors()
        if not openedFiles:
            # No opened files, fallback to search in dir
            self.setSearchInDirectory(what, None)
            return

        # Select the radio buttons
        self.projectRButton.setEnabled(GlobalData().project.isLoaded())
        self.openFilesRButton.setEnabled(True)
        self.openFilesRButton.setChecked(True)
        self.dirEditCombo.setEnabled(False)
        self.dirSelectButton.setEnabled(False)

        if what:
            # Pick up the history values if so
            comboIndex, historyIndex = self.__historyIndexByWhat(what)
            if historyIndex is not None:
                self.__deserialize(self.__history[historyIndex])
                self.findCombo.setCurrentIndex(comboIndex)
            else:
                self.findCombo.setCurrentText(what)
            self.findCombo.lineEdit().selectAll()
        self.findCombo.setFocus()

        # Check searchability
        self.__testSearchability()

    def setSearchInDirectory(self, what=None, dirPath=None):
        """Sets search ready for the given directory"""
        # Select radio buttons
        self.projectRButton.setEnabled(GlobalData().project.isLoaded())
        openedFiles = self.editorsManager.getTextEditors()
        self.openFilesRButton.setEnabled(len(openedFiles) != 0)
        self.dirRButton.setEnabled(True)
        self.dirRButton.setChecked(True)
        self.dirEditCombo.setEnabled(True)
        self.dirSelectButton.setEnabled(True)

        if what:
            # Pick up the history values if so
            comboIndex, historyIndex = self.__historyIndexByWhat(what)
            if historyIndex is not None:
                self.__deserialize(self.__history[historyIndex])
                self.findCombo.setCurrentIndex(comboIndex)
            else:
                self.findCombo.setCurrentText(what)
            self.findCombo.lineEdit().selectAll()

        if dirPath:
            self.dirEditCombo.setEditText(dirPath)

        self.findCombo.setFocus()

        # Check searchability
        self.__testSearchability()

    @staticmethod
    def __normalizeFilters(text):
        """Normalizes the filters string"""
        normParts = []
        for part in text.strip().split(';'):
            part = part.strip()
            if part:
                normParts.append(part)
        return '; '.join(normParts)

    def __testSearchability(self):
        """Tests the searchability and sets the Find button status"""
        startTime = time.time()
        if self.findCombo.currentText().strip() == "":
            self.findButton.setEnabled(False)
            self.findButton.setToolTip("No text to search")
            return

        if self.dirRButton.isChecked():
            dirname = self.dirEditCombo.currentText().strip()
            if dirname == "":
                self.findButton.setEnabled(False)
                self.findButton.setToolTip("No directory path")
                return
            if not isdir(dirname):
                self.findButton.setEnabled(False)
                self.findButton.setToolTip("Path is not a directory")
                return

        # Now we need to match file names if there is a filter
        filtersText = self.filterCombo.currentText().strip()
        if filtersText == "":
            self.findButton.setEnabled(True)
            self.findButton.setToolTip("Find in files")
            return

        # Need to check the files match
        try:
            filters = self.__compileFilters()
        except:
            self.findButton.setEnabled(False)
            self.findButton.setToolTip("Incorrect files "
                                       "filter regular expression")
            return

        matched = False
        tooLong = False
        if self.projectRButton.isChecked():
            # Whole project
            for fname in GlobalData().project.filesList:
                if fname.endswith(sep):
                    continue
                matched = self.__filterMatch(filters, fname)
                if matched:
                    break
                # Check the time, it might took too long
                if time.time() - startTime > 0.1:
                    tooLong = True
                    break

        elif self.openFilesRButton.isChecked():
            # Opened files
            openedFiles = self.editorsManager.getTextEditors()
            for record in openedFiles:
                matched = self.__filterMatch(filters, record[1])
                if matched:
                    break
                # Check the time, it might took too long
                if time.time() - startTime > 0.1:
                    tooLong = True
                    break

        else:
            # Search in the dir
            if not dirname.endswith(sep):
                dirname += sep
            matched, tooLong = self.__matchInDir(dirname, filters, startTime)

        if matched:
            self.findButton.setEnabled(True)
            self.findButton.setToolTip("Find in files")
        else:
            if tooLong:
                self.findButton.setEnabled(True)
                self.findButton.setToolTip("Find in files")
            else:
                self.findButton.setEnabled(False)
                self.findButton.setToolTip("No files matched to search in")

    @staticmethod
    def __matchInDir(path, filters, startTime):
        """Provides the 'match' and 'too long' statuses"""
        matched = False
        tooLong = False
        for item in listdir(path):
            if time.time() - startTime > 0.1:
                tooLong = True
                return matched, tooLong
            if isdir(path + item):
                dname = path + item + sep
                matched, tooLong = FindInFilesDialog.__matchInDir(dname,
                                                                  filters,
                                                                  startTime)
                if matched or tooLong:
                    return matched, tooLong
                continue
            if FindInFilesDialog.__filterMatch(filters, path + item):
                matched = True
                return matched, tooLong
        return matched, tooLong

    def __projectClicked(self):
        """project radio button clicked"""
        self.dirEditCombo.setEnabled(False)
        self.dirSelectButton.setEnabled(False)
        self.__testSearchability()

    def __openFilesOnlyClicked(self):
        """open files only radio button clicked"""
        self.dirEditCombo.setEnabled(False)
        self.dirSelectButton.setEnabled(False)
        self.__testSearchability()

    def __dirClicked(self):
        """dir radio button clicked"""
        self.dirEditCombo.setEnabled(True)
        self.dirSelectButton.setEnabled(True)
        self.dirEditCombo.setFocus()
        self.__testSearchability()

    def __someTextChanged(self, text):
        """Text to search, filter or dir name has been changed"""
        del text    # unused argument
        self.__testSearchability()

    def __whatIndexChanged(self, index):
        """Index in history has changed"""
        if index != -1:
            historyIndex = self.findCombo.itemData(index)
            if historyIndex is not None:
                self.__deserialize(self.__history[historyIndex])
        self.__testSearchability()

    def __selectDirClicked(self):
        """The user selects a directory"""
        options = QFileDialog.Options()
        options |= QFileDialog.ShowDirsOnly | QFileDialog.DontUseNativeDialog
        dirName = QFileDialog.getExistingDirectory(
            self, 'Select directory to search in',
            self.dirEditCombo.currentText(), options)

        if dirName:
            self.dirEditCombo.setEditText(normpath(dirName))
        self.__testSearchability()

    @staticmethod
    def __filterMatch(filters, fname):
        """True if the file should be taken into consideration"""
        if filters:
            for filt in filters:
                if filt.match(fname):
                    return True
            return False
        return True

    def __projectFiles(self, filters):
        """Project files list respecting the mask"""
        mainWindow = GlobalData().mainWindow
        files = []
        for fname in GlobalData().project.filesList:
            if fname.endswith(sep):
                continue
            if self.__filterMatch(filters, fname):
                widget = mainWindow.getWidgetForFileName(fname)
                if widget is None:
                    # Do not check for broken symlinks
                    if isFileSearchable(fname, False):
                        files.append(ItemToSearchIn(fname, ""))
                else:
                    if widget.getType() in \
                                [MainWindowTabWidgetBase.PlainTextEditor]:
                        files.append(ItemToSearchIn(fname,
                                                    widget.getUUID()))
            QApplication.processEvents()
            if self.__cancelRequest:
                raise Exception("Cancel request")
        return files

    def __openedFiles(self, filters):
        """Currently opened editor buffers"""
        files = []
        openedFiles = self.editorsManager.getTextEditors()
        for record in openedFiles:
            uuid = record[0]
            fname = record[1]
            if self.__filterMatch(filters, fname):
                files.append(ItemToSearchIn(fname, uuid))
            QApplication.processEvents()
            if self.__cancelRequest:
                raise Exception("Cancel request")
        return files

    def __dirFiles(self, path, filters, files):
        """Files recursively for the dir"""
        for item in listdir(path):
            QApplication.processEvents()
            if self.__cancelRequest:
                raise Exception("Cancel request")
            if isdir(path + item):
                if item in ['.svn', '.cvs', '.git', '.hg']:
                    # It does not make sense to search in revision control dirs
                    continue
                anotherDir, isLoop = resolveLink(path + item)
                if not isLoop:
                    self.__dirFiles(anotherDir + sep,
                                    filters, files)
                continue
            if not isfile(path + item):
                continue
            realItem, isLoop = resolveLink(path + item)
            if isLoop:
                continue
            if self.__filterMatch(filters, realItem):
                found = False
                for itm in files:
                    if itm.fileName == realItem:
                        found = True
                        break
                if not found:
                    mainWindow = GlobalData().mainWindow
                    widget = mainWindow.getWidgetForFileName(realItem)
                    if widget is None:
                        if isFileSearchable(realItem):
                            files.append(ItemToSearchIn(realItem, ""))
                    else:
                        if widget.getType() in \
                                    [MainWindowTabWidgetBase.PlainTextEditor]:
                            files.append(ItemToSearchIn(realItem,
                                                        widget.getUUID()))

    def __compileFilters(self):
        """Compiles the filters"""
        filtersText = self.filterCombo.currentText().strip()
        filtersRe = []
        if filtersText != "":
            for filt in filtersText.split(';'):
                filtersRe.append(re.compile(filt.strip(), re.IGNORECASE))
        return filtersRe

    def __buildFilesList(self):
        """Builds the list of files to search in"""
        filtersRe = self.__compileFilters()

        if self.projectRButton.isChecked():
            return self.__projectFiles(filtersRe)

        if self.openFilesRButton.isChecked():
            return self.__openedFiles(filtersRe)

        dirname = realpath(self.dirEditCombo.currentText().strip())
        files = []
        self.__dirFiles(dirname + sep, filtersRe, files)
        return files

    def __updateHistory(self):
        """Updates history if needed"""
        # Add entries to the combo box if required
        historyItem = self.__serialize()
        _, historyIndex = self.__historyIndexByWhat(
            self.findCombo.currentText())
        if historyIndex is not None:
            self.__history[historyIndex] = historyItem
        else:
            historyIndex = 0
            self.__history.insert(0, historyItem)
            if len(self.__history) > self.__maxEntries:
                self.__history = self.__history[:self.__maxEntries]

        self.findCombo.clear()
        self.filterCombo.clear()
        self.dirEditCombo.clear()
        self.__populateHistory()

        self.findCombo.setCurrentIndex(historyIndex)
        self.findCombo.setCurrentText(historyItem['term'])

        fltValue = historyItem['filters']
        if fltValue:
            index = self.__indexByText(self.filterCombo, fltValue)
            self.filterCombo.setCurrentIndex(index)
        self.filterCombo.setCurrentText(fltValue)

        dirValue = historyItem['dir']
        if dirValue:
            index = self.__indexByText(self.dirEditCombo, dirValue)
            self.dirEditCombo.setCurrentIndex(index)
        self.dirEditCombo.setCurrentText(dirValue)

        # Save the combo values for further usage
        setFindInFilesHistory(self.__history)

    def __process(self):
        """Search process"""
        self.__updateHistory()

        self.__inProgress = True
        numberOfMatches = 0
        self.searchResults = []
        self.searchRegexp = None

        # Form the regexp to search
        regexpText = self.findCombo.currentText()
        if not self.regexpCheckBox.isChecked():
            regexpText = re.escape(regexpText)
        if self.wordCheckBox.isChecked():
            regexpText = "\\b%s\\b" % regexpText
        flags = re.UNICODE
        if not self.caseCheckBox.isChecked():
            flags |= re.IGNORECASE

        try:
            self.searchRegexp = re.compile(regexpText, flags)
        except Exception as exc:
            logging.error("Invalid search expression: %s", str(exc))
            self.close()
            return

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        self.fileLabel.setPath('Building list of files to search in...')
        QApplication.processEvents()
        try:
            files = self.__buildFilesList()
        except Exception as exc:
            QApplication.restoreOverrideCursor()
            if 'Cancel request' not in str(exc):
                logging.error(str(exc))
            self.close()
            return
        QApplication.restoreOverrideCursor()
        QApplication.processEvents()

        if not files:
            self.fileLabel.setPath('No files to search in')
            return

        self.progressBar.setRange(0, len(files))

        index = 1
        for item in files:

            if self.__cancelRequest:
                self.__inProgress = False
                self.close()
                return

            self.fileLabel.setPath('Matches: ' + str(numberOfMatches) +
                                   ' Processing: ' + item.fileName)

            item.search(self.searchRegexp)
            found = len(item.matches)
            if found > 0:
                numberOfMatches += found
                self.searchResults.append(item)

            self.progressBar.setValue(index)
            index += 1

            QApplication.processEvents()

        if numberOfMatches == 0:
            msg = 'No matches in ' + str(len(files)) + 'file'
            if len(files) > 1:
                msg += 's'
            self.fileLabel.setPath(msg)
            self.__inProgress = False
        else:
            self.close()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""imports diagram dialog"""

import os
import os.path
import logging
from ui.qt import (Qt, QTimer, QDialog, QDialogButtonBox, QVBoxLayout, QLabel,
                   QCheckBox, QProgressBar, QApplication, QGraphicsScene,
                   QGraphicsPixmapItem)
from utils.globals import GlobalData
from utils.fileutils import isPythonFile
from utils.importutils import resolveImports
from utils.pixmapcache import getPixmap
from cdmpyparser import getBriefModuleInfoFromMemory
from .plaindotparser import getGraphFromDescriptionData
from .importsdgmgraphics import (ImportsDgmDocConn, ImportsDgmDependConn,
                                 ImportsDgmSystemWideModule,
                                 ImportsDgmUnknownModule,
                                 ImportsDgmBuiltInModule,
                                 ImportsDgmModuleOfInterest,
                                 ImportsDgmOtherPrjModule,
                                 ImportsDgmDocNote, ImportsDgmEdgeLabel)


class DgmConnection:

    """Holds information about one connection"""

    # Connection types
    ModuleDoc = 0
    ModuleDependency = 1

    def __init__(self):
        self.objName = ""       # Unique object name
        self.kind = -1          # See connection types
        self.source = ""        # Connection start point
        self.target = ""        # Connection end point
        self.labels = []        # Connection labels: list of what imported

    def toGraphviz(self):
        """Serialize the connection in graphviz format"""
        attributes = 'id="' + self.objName + '", arrowhead=none'
        label = ""
        for what in self.labels:
            if label != "":
                label += "\\n"
            label += what
        if label != "":
            attributes += ', label="' + label + \
                          '", fontname=Arial, fontsize=10'

        return self.source + " -> " + self.target + '[ ' + attributes + ' ];'

    def __eq__(self, other):
        """Checks if the connection connects the same objects"""
        return self.source == other.source and self.target == other.target


class DgmDocstring:

    """Holds information about one docstring"""

    def __init__(self):
        self.objName = ""       # Unique object name
        self.docstring = None   # Module docstring object

        self.refFile = ""       # File of the docstring

    def toGraphviz(self):
        """Serialize the docstring box in graphviz format"""
        escapedText = self.docstring.text.replace('\n', '\\n')
        escapedText = escapedText.replace('"', '\\"')
        attributes = 'shape=box, fontname=Arial, fontsize=10'
        return self.objName + \
               ' [ ' + attributes + ', label="' + escapedText + '" ];'


class DgmModule:

    """Holds information about one module"""

    # Module types
    ModuleOfInterest = 0
    OtherProjectModule = 1
    SystemWideModule = 2
    BuiltInModule = 3
    UnknownModule = 4

    def __init__(self):
        self.objName = ""       # Unique object name
        self.kind = -1          # See module types
        self.title = ""         # title
        self.classes = []       # list of classes objects
        self.funcs = []         # list of funcs objects
        self.globs = []         # list of global var objects
        self.imports = []       # list of imports

        self.refFile = ""       # File of the module
        self.docstring = ""

    def toGraphviz(self):
        """Serialize the module box in graphviz format"""
        classesPart = ""
        funcsPart = ""
        globsPart = ""

        for klass in self.classes:
            if classesPart != "":
                classesPart += "\\n"
            classesPart += klass.name
        for func in self.funcs:
            if funcsPart != "":
                funcsPart += "\\n"
            funcsPart += func.name
        for glob in self.globs:
            if globsPart != "":
                globsPart += "\\n"
            globsPart += glob.name

        spareForTopBottom = "\\n"

        attributes = 'shape=box, fontname=Arial, fontsize=10'

        if self.isProjectModule():
            return self.objName + ' [ ' + attributes + \
                   ', label="' + spareForTopBottom + self.title + '\\n' + \
                   classesPart + '\\n' + funcsPart + '\\n' + globsPart + '" ];'
        return self.objName + ' [ ' + attributes + \
               ', label="' + self.title + '" ];'

    def isProjectModule(self):
        """True if belongs to the project or the dir of interest"""
        return self.kind in [self.ModuleOfInterest, self.OtherProjectModule]

    def __eq__(self, other):
        """Compares two module boxes when they are added to the data model"""
        if self.isProjectModule() and other.isProjectModule():
            return self.refFile == other.refFile
        return self.refFile == other.refFile and \
               self.kind == other.kind and self.title == other.title

    def getTooltip(self):
        """Provides a tooltip"""
        tooltip = ''
        if self.refFile != "":
            tooltip = self.refFile
        if self.docstring != "":
            if tooltip != "":
                tooltip += "\n\n"
            tooltip += self.docstring
        return tooltip


class DgmRank:

    """Holds information about one rank"""

    def __init__(self):
        self.firstObj = ""
        self.secondObj = ""

    def __eq__(self, other):
        """Compares two ranks"""
        return self.firstObj == other.firstObj and \
               self.secondObj == other.secondObj

    def toGraphviz(self):
        """Serialize the rank in graphviz format"""
        return '{ rank=same; "' + self.firstObj + '"; "' + \
               self.secondObj + '"; }'


class ImportDiagramModel:

    """Holds information about data model of an import diagram"""

    def __init__(self):
        self.modules = []
        self.docstrings = []
        self.connections = []
        self.ranks = []

        self.__objectsCounter = -1

    def clear(self):
        """Clears the diagram model"""
        self.modules = []
        self.docstrings = []
        self.connections = []
        self.ranks = []

        self.__objectsCounter = -1

    def toGraphviz(self):
        """Serialize the import diagram in graphviz format"""
        result = "digraph ImportsDiagram { "
        for item in self.docstrings:
            result += item.toGraphviz() + "\n"
        for item in self.modules:
            result += item.toGraphviz() + "\n"
        for item in self.connections:
            result += item.toGraphviz() + "\n"
        for item in self.ranks:
            result += item.toGraphviz() + "\n"
        result += "}"
        return result

    def __newName(self):
        """Generates a short name for the graphviz objects"""
        self.__objectsCounter += 1
        return "obj" + str(self.__objectsCounter)

    def addRank(self, rank):
        """Adds a rank"""
        for idx in range(0, len(self.ranks)):
            if self.ranks[idx] == rank:
                return
        self.ranks.append(rank)

    def addConnection(self, conn):
        """Adds a connection and provides its name"""
        # The connections can be added twice if there are two import directives
        index = -1
        for idx in range(0, len(self.connections)):
            if self.connections[idx] == conn:
                index = idx
                break

        if index == -1:
            # new connection, generate name and add
            conn.objName = self.__newName()
            self.connections.append(conn)
            return conn.objName

        # There is already such a connection. So merge labels.
        self.connections[index].labels += conn.labels
        return self.connections[index].objName

    def addDocstringBox(self, docBox):
        """Adds a module docstring"""
        # Docstring boxes cannot appear twice so just add it
        docBox.objName = self.__newName()
        self.docstrings.append(docBox)
        return docBox.objName

    def addModule(self, modBox):
        """Adds a module box"""
        # It might happened that the same module appeared more than once
        index = -1
        for idx in range(0, len(self.modules)):
            if self.modules[idx] == modBox:
                index = idx
                break

        if index == -1:
            # New module box, generate name and add it
            modBox.objName = self.__newName()
            self.modules.append(modBox)
            return modBox.objName

        # There is already such a box, so the box type might need to be
        # adjusted
        if modBox.kind == self.modules[index].kind:
            return self.modules[index].objName

        if modBox.kind == DgmModule.ModuleOfInterest:
            # Replace the existed one with a new one but keep the object name
            modBox.objName = self.modules[index].objName
            self.modules[index] = modBox
            return modBox.objName

        # No need in adjustments
        return self.modules[index].objName

    def findModule(self, name):
        """Searches for a module by the object name"""
        for obj in self.modules:
            if obj.objName == name:
                return obj
        return None

    def findConnection(self, name, tail=""):
        """Searches for a connection by the object name"""
        if tail == "":
            # Search by the connection name
            for obj in self.connections:
                if obj.objName == name:
                    return obj
            return None

        # Search by the name of the objects it connects
        for obj in self.connections:
            if obj.source == name and obj.target == tail:
                return obj
        return None

    def findDocstring(self, name):
        """Searches for a docstring by the object name"""
        for obj in self.docstrings:
            if obj.objName == name:
                return obj
        return None


class ImportDiagramOptions:

    """Holds the generated diagram settings"""

    def __init__(self):
        self.includeClasses = True
        self.includeFuncs = True
        self.includeGlobs = True
        self.includeDocs = False
        self.includeConnText = True


class ImportsDiagramDialog(QDialog):

    """Imports diagram properties dialog implementation"""

    # Options of providing a diagram
    SingleFile = 0
    DirectoryFiles = 1
    ProjectFiles = 2
    SingleBuffer = 3

    def __init__(self, option, path="", parent=None):
        QDialog.__init__(self, parent)

        self.__cancelRequest = False
        self.__inProgress = False
        self.__option = option
        self.__path = path

        # Avoid pylint complains
        self.includeClassesBox = None
        self.includeFuncsBox = None
        self.includeGlobsBox = None
        self.includeDocsBox = None
        self.includeConnTextBox = None

        self.options = ImportDiagramOptions()

        self.__createLayout()
        title = "Imports diagram settings for "
        if self.__option == self.SingleFile:
            title += os.path.basename(self.__path)
        elif self.__option == self.DirectoryFiles:
            title += "directory " + self.__path
        elif self.__option == self.ProjectFiles:
            title += "the whole project"
        else:
            title += "modified file " + os.path.basename(self.__path)
        self.setWindowTitle(title)

    def __updateOptions(self, state=None):
        """Updates the saved options"""
        self.options.includeClasses = self.includeClassesBox.isChecked()
        self.options.includeFuncs = self.includeFuncsBox.isChecked()
        self.options.includeGlobs = self.includeGlobsBox.isChecked()
        self.options.includeDocs = self.includeDocsBox.isChecked()
        self.options.includeConnText = self.includeConnTextBox.isChecked()

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(400, 100)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)

        # Check boxes
        self.includeClassesBox = QCheckBox(self)
        self.includeClassesBox.setText("Show &classes in modules")
        self.includeClassesBox.setChecked(self.options.includeClasses)
        self.includeClassesBox.stateChanged.connect(self.__updateOptions)
        self.includeFuncsBox = QCheckBox(self)
        self.includeFuncsBox.setText("Show &functions in modules")
        self.includeFuncsBox.setChecked(self.options.includeFuncs)
        self.includeFuncsBox.stateChanged.connect(self.__updateOptions)
        self.includeGlobsBox = QCheckBox(self)
        self.includeGlobsBox.setText("Show &global variables in modules")
        self.includeGlobsBox.setChecked(self.options.includeGlobs)
        self.includeGlobsBox.stateChanged.connect(self.__updateOptions)
        self.includeDocsBox = QCheckBox(self)
        self.includeDocsBox.setText("Show modules &docstrings")
        self.includeDocsBox.setChecked(self.options.includeDocs)
        self.includeDocsBox.stateChanged.connect(self.__updateOptions)
        self.includeConnTextBox = QCheckBox(self)
        self.includeConnTextBox.setText("Show connection &labels")
        self.includeConnTextBox.setChecked(self.options.includeConnText)
        self.includeConnTextBox.stateChanged.connect(self.__updateOptions)

        verticalLayout.addWidget(self.includeClassesBox)
        verticalLayout.addWidget(self.includeFuncsBox)
        verticalLayout.addWidget(self.includeGlobsBox)
        verticalLayout.addWidget(self.includeDocsBox)
        verticalLayout.addWidget(self.includeConnTextBox)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Cancel)
        generateButton = buttonBox.addButton("Generate",
                                             QDialogButtonBox.ActionRole)
        generateButton.setDefault(True)
        generateButton.clicked.connect(self.accept)
        verticalLayout.addWidget(buttonBox)

        buttonBox.rejected.connect(self.close)


class ImportsDiagramProgress(QDialog):

    """Progress of the diagram generator"""

    def __init__(self, what, options, path="", buf="", parent=None):
        QDialog.__init__(self, parent)
        self.__cancelRequest = False
        self.__inProgress = False

        self.__what = what
        self.__options = options
        self.__path = path          # could be a dir or a file
        self.__buf = buf            # content in case of a modified file

        # Working process data
        self.__participantFiles = []    # Collected list of files
        self.__projectImportDirs = []
        self.__projectImportsCache = {} # utils.settings -> /full/path/to.py
        self.__dirsToImportsCache = {}  # /dir/path -> { my.mod: path.py, ... }

        self.dataModel = ImportDiagramModel()
        self.scene = QGraphicsScene()

        # Avoid pylint complains
        self.progressBar = None
        self.infoLabel = None

        self.__createLayout()
        self.setWindowTitle('Imports/dependencies diagram generator')
        QTimer.singleShot(0, self.__process)

    def keyPressEvent(self, event):
        """Processes the ESC key specifically"""
        if event.key() == Qt.Key_Escape:
            self.__onClose()
        else:
            QDialog.keyPressEvent(self, event)

    def __createLayout(self):
        """Creates the dialog layout"""
        self.resize(450, 20)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)

        # Info label
        self.infoLabel = QLabel(self)
        verticalLayout.addWidget(self.infoLabel)

        # Progress bar
        self.progressBar = QProgressBar(self)
        self.progressBar.setValue(0)
        self.progressBar.setOrientation(Qt.Horizontal)
        verticalLayout.addWidget(self.progressBar)

        # Buttons
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Close)
        verticalLayout.addWidget(buttonBox)

        buttonBox.rejected.connect(self.__onClose)

    def __onClose(self):
        """triggered when the close button is clicked"""
        self.__cancelRequest = True
        if not self.__inProgress:
            self.close()

    def __buildParticipants(self):
        """Builds a list of participating files and dirs"""
        if self.__what in [ImportsDiagramDialog.SingleBuffer,
                           ImportsDiagramDialog.SingleFile]:
            # File exists but could be modified
            self.__path = os.path.realpath(self.__path)
            self.__participantFiles.append(self.__path)
            return

        if self.__what == ImportsDiagramDialog.ProjectFiles:
            self.__scanProjectDirs()
            return

        # This is a recursive directory
        self.__path = os.path.realpath(self.__path)
        self.__scanDirForPythonFiles(self.__path + os.path.sep)

    def __scanDirForPythonFiles(self, path):
        """Scans the directory for the python files recursively"""
        for item in os.listdir(path):
            if item in [".svn", ".cvs", '.git', '.hg']:
                continue
            if os.path.isdir(path + item):
                self.__scanDirForPythonFiles(path + item + os.path.sep)
                continue
            if isPythonFile(path + item):
                self.__participantFiles.append(os.path.realpath(path + item))

    def __scanProjectDirs(self):
        """Populates participant lists from the project files"""
        for fName in GlobalData().project.filesList:
            if isPythonFile(fName):
                self.__participantFiles.append(fName)

    def __addBoxInfo(self, box, info):
        """Adds information to the given box if so configured"""
        if info.docstring is not None:
            box.docstring = info.docstring.text

        if self.__options.includeClasses:
            for klass in info.classes:
                box.classes.append(klass)

        if self.__options.includeFuncs:
            for func in info.functions:
                box.funcs.append(func)

        if self.__options.includeGlobs:
            for glob in info.globals:
                box.globs.append(glob)

        if self.__options.includeConnText:
            for imp in info.imports:
                box.imports.append(imp)

    def __addDocstringBox(self, info, fName, modBoxName):
        """Adds a docstring box if needed"""
        if self.__options.includeDocs:
            if info.docstring is not None:
                docBox = DgmDocstring()
                docBox.docstring = info.docstring
                docBox.refFile = fName

                # Add the box and its connection
                docBoxName = self.dataModel.addDocstringBox(docBox)

                conn = DgmConnection()
                conn.kind = DgmConnection.ModuleDoc
                conn.source = modBoxName
                conn.target = docBoxName
                self.dataModel.addConnection(conn)

                # Add rank for better layout
                rank = DgmRank()
                rank.firstObj = modBoxName
                rank.secondObj = docBoxName
                self.dataModel.addRank(rank)

    def __getSytemWideImportDocstring(self, path):
        """Provides the system wide module docstring"""
        if isPythonFile(path):
            try:
                info = GlobalData().briefModinfoCache.get(path)
                if info.docstring is not None:
                    return info.docstring.text
            except:
                pass
        return ''

    @staticmethod
    def __getModuleTitle(fName):
        """Extracts a module name out of the file name"""
        baseTitle = os.path.basename(fName).split('.')[0]
        if baseTitle != "__init__":
            return baseTitle

        # __init__ is not very descriptive. Add a top level dir.
        dirName = os.path.dirname(fName)
        topDir = os.path.basename(dirName)
        return topDir + "(" + baseTitle + ")"

    @staticmethod
    def __isLocalOrProject(fName, resolvedPath):
        """True if the module is a project one or is in the nested dirs"""
        if resolvedPath is None:
            return False
        if not os.path.isabs(resolvedPath):
            return False
        if GlobalData().project.isProjectFile(resolvedPath):
            return True

        resolvedDir = os.path.dirname(resolvedPath)
        baseDir = os.path.dirname(fName)
        return resolvedDir.startswith(baseDir)

    def __addSingleFileToDataModel(self, info, fName):
        """Adds a single file to the data model"""
        if fName.endswith('__init__.py'):
            if not info.classes and not info.functions and \
               not info.globals and not info.imports:
                # Skip dummy init files
                return

        modBox = DgmModule()
        modBox.refFile = fName

        modBox.kind = DgmModule.ModuleOfInterest
        modBox.title = self.__getModuleTitle(fName)

        self.__addBoxInfo(modBox, info)
        modBoxName = self.dataModel.addModule(modBox)
        self.__addDocstringBox(info, fName, modBoxName)

        # Analyze what was imported
        resolvedImports, errors = resolveImports(fName, info.imports)
        if errors:
            message = 'Errors while analyzing ' + fName + ':'
            for err in errors:
                message += '\n    ' + err
            logging.warning(message)

        for item in resolvedImports:
            importName = item[0]        # from name
            resolvedPath = item[1]      # 'built-in', None or absolute path
            importedNames = item[2]     # list of strings

            impBox = DgmModule()
            impBox.title = importName

            if self.__isLocalOrProject(fName, resolvedPath):
                impBox.kind = DgmModule.OtherProjectModule
                impBox.refFile = resolvedPath
                if isPythonFile(resolvedPath):
                    otherInfo = GlobalData().briefModinfoCache.get(
                        resolvedPath)
                    self.__addBoxInfo(impBox, otherInfo)
            else:
                if resolvedPath is None:
                    # e.g. 'import sys' will have None for the path
                    impBox.kind = DgmModule.UnknownModule
                elif os.path.isabs(resolvedPath):
                    impBox.kind = DgmModule.SystemWideModule
                    impBox.refFile = resolvedPath
                    impBox.docstring = \
                        self.__getSytemWideImportDocstring(resolvedPath)
                else:
                    # e.g. 'import time' will have 'built-in' in the path
                    impBox.kind = DgmModule.BuiltInModule

            impBoxName = self.dataModel.addModule(impBox)

            impConn = DgmConnection()
            impConn.kind = DgmConnection.ModuleDependency
            impConn.source = modBoxName
            impConn.target = impBoxName

            if self.__options.includeConnText:
                for impWhat in importedNames:
                    if impWhat:
                        impConn.labels.append(impWhat)
            self.dataModel.addConnection(impConn)

    def __process(self):
        """Accumulation process"""
        # Intermediate working data
        self.__participantFiles = []
        self.__projectImportDirs = []
        self.__projectImportsCache = {}

        self.dataModel.clear()
        self.__inProgress = True

        try:
            self.infoLabel.setText('Building the list of files to analyze...')
            QApplication.processEvents()

            # Build the list of participating python files
            self.__buildParticipants()
            self.__projectImportDirs = \
                GlobalData().project.getImportDirsAsAbsolutePaths()


            QApplication.processEvents()
            if self.__cancelRequest:
                QApplication.restoreOverrideCursor()
                self.close()
                return

            self.progressBar.setRange(0, len(self.__participantFiles))
            index = 1

            # Now, parse the files and build the diagram data model
            if self.__what == ImportsDiagramDialog.SingleBuffer:
                info = getBriefModuleInfoFromMemory(str(self.__buf))
                self.__addSingleFileToDataModel(info, self.__path)
            else:
                infoSrc = GlobalData().briefModinfoCache
                for fName in self.__participantFiles:
                    self.progressBar.setValue(index)
                    self.infoLabel.setText('Analyzing ' + fName + "...")
                    QApplication.processEvents()
                    if self.__cancelRequest:
                        QApplication.restoreOverrideCursor()
                        self.dataModel.clear()
                        self.close()
                        return
                    info = infoSrc.get(fName)
                    self.__addSingleFileToDataModel(info, fName)
                    index += 1

            # The import caches and other working data are not needed anymore
            self.__participantFiles = None
            self.__projectImportDirs = None
            self.__projectImportsCache = None

            # Generating the graphviz layout
            self.infoLabel.setText('Generating layout using graphviz...')
            QApplication.processEvents()

            graph = getGraphFromDescriptionData(self.dataModel.toGraphviz())
            graph.normalize(self.physicalDpiX(), self.physicalDpiY())
            QApplication.processEvents()
            if self.__cancelRequest:
                QApplication.restoreOverrideCursor()
                self.dataModel.clear()
                self.close()
                return

            # Generate graphics scene
            self.infoLabel.setText('Generating graphics scene...')
            QApplication.processEvents()
            self.__buildGraphicsScene(graph)

            # Clear the data model
            self.dataModel = None
        except Exception as exc:
            QApplication.restoreOverrideCursor()
            logging.error(str(exc))
            self.__inProgress = False
            self.__onClose()
            return

        QApplication.restoreOverrideCursor()
        self.infoLabel.setText('Done')
        QApplication.processEvents()
        self.__inProgress = False

        self.accept()

    def __buildGraphicsScene(self, graph):
        """Builds the QT graphics scene"""
        self.scene.clear()
        self.scene.setSceneRect(0, 0, graph.width, graph.height)

        for edge in graph.edges:
            # self.scene.addItem( GraphicsEdge( edge, self ) )
            dataModelObj = self.dataModel.findConnection(edge.tail, edge.head)
            if dataModelObj is None:
                raise Exception("Cannot find the following connection: " +
                                edge.tail + " -> " + edge.head)

            if dataModelObj.kind == DgmConnection.ModuleDoc:
                modObj = self.dataModel.findModule(dataModelObj.source)
                if modObj is None:
                    raise Exception("Cannot find module object: " +
                                    dataModelObj.source)
                self.scene.addItem(ImportsDgmDocConn(edge, modObj))
                continue
            if dataModelObj.kind == DgmConnection.ModuleDependency:
                # Find the source module object first
                modObj = self.dataModel.findModule(dataModelObj.source)
                if modObj is None:
                    raise Exception("Cannot find module object: " +
                                    dataModelObj.source)
                self.scene.addItem(
                    ImportsDgmDependConn(edge, modObj, dataModelObj))

                if edge.label != "":
                    self.scene.addItem(ImportsDgmEdgeLabel(edge, modObj))
                continue

            raise Exception("Unexpected type of connection: " +
                            str(dataModelObj.kind))

        for node in graph.nodes:
            dataModelObj = self.dataModel.findModule(node.name)
            if dataModelObj is None:
                dataModelObj = self.dataModel.findDocstring(node.name)
            if dataModelObj is None:
                raise Exception("Cannot find object " + node.name)

            if isinstance(dataModelObj, DgmDocstring):
                self.scene.addItem(ImportsDgmDocNote(node,
                                                     dataModelObj.refFile,
                                                     dataModelObj.docstring))
                continue

            # OK, this is a module rectangle. Switch by type of the module.
            if dataModelObj.kind == DgmModule.ModuleOfInterest:
                self.scene.addItem(
                    ImportsDgmModuleOfInterest(node, dataModelObj.refFile,
                                               dataModelObj,
                                               self.physicalDpiX()))
            elif dataModelObj.kind == DgmModule.OtherProjectModule:
                self.scene.addItem(
                    ImportsDgmOtherPrjModule(node, dataModelObj.refFile,
                                             dataModelObj,
                                             self.physicalDpiX()))
            elif dataModelObj.kind == DgmModule.SystemWideModule:
                self.scene.addItem(
                    ImportsDgmSystemWideModule(node,
                                               dataModelObj.refFile,
                                               dataModelObj.docstring))
            elif dataModelObj.kind == DgmModule.BuiltInModule:
                self.scene.addItem(ImportsDgmBuiltInModule(node))
            elif dataModelObj.kind == DgmModule.UnknownModule:
                self.scene.addItem(ImportsDgmUnknownModule(node))
            else:
                raise Exception("Unexpected type of module: " +
                                str(dataModelObj.kind))

            tooltip = dataModelObj.getTooltip()
            if tooltip:
                pixmap = getPixmap('diagramdoc.png')
                docItem = QGraphicsPixmapItem(pixmap)
                docItem.setToolTip(tooltip)
                posX = node.posX + node.width / 2.0 - pixmap.width() / 2.0
                posY = node.posY - node.height / 2.0 - pixmap.height() / 2.0
                docItem.setPos(posX, posY)
                self.scene.addItem(docItem)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Quick help screen"""

from flowui.cml import CMLVersion
from .texttabwidget import TextTabWidget


class QuickHelpWidget(TextTabWidget):

    """Quick help screen"""

    def __init__(self, parent=None):

        TextTabWidget.__init__(self, parent)
        # pixmapPath = os.path.dirname( os.path.abspath( sys.argv[0] ) ) + \
        #              os.path.sep + 'pixmaps' + os.path.sep
        # logoPath = pixmapPath + 'logo.png'

        self.setHTML("""
<html>
<div>

    <h2 align="center">Keyboard Shortcut Reference</h2>

    <h3>Tools</h3>
    <p align="center">
      <table border="0" cellspacing="1"
             cellpadding="4" width="95%" align="center">
        <tr>
          <td bgcolor="#E9E9F3">Alt+Shift+S</td>
          <td bgcolor="#F6F4E4">Search a name</td>
          <td bgcolor="#E9E9F3">Alt+Shift+O</td>
          <td bgcolor="#F6F4E4">Search a file</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+I</td>
          <td bgcolor="#F6F4E4">Open import / select import to open</td>
          <td bgcolor="#E9E9F3"></td>
          <td bgcolor="#F6F4E4"></td>
        </tr>
      </table>
    </p>

    <br>
    <h3>IDE</h3>
    <p align="center">
      <table border="0" cellspacing="1"
             cellpadding="4" width="95%" align="center">
        <tr>
          <td width="15%" bgcolor="#E9E9F3">Ctrl+N</td>
          <td width="35%" bgcolor="#F6F4E4">New file</td>
          <td width="15%" bgcolor="#E9E9F3">F11</td>
          <td width="35%" bgcolor="#F6F4E4">Shrink sidebars</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Alt+PgUp or Down</td>
          <td bgcolor="#F6F4E4">Forward or back in editing history</td>
          <td bgcolor="#E9E9F3">Ctrl+PgUp or Down</td>
          <td bgcolor="#F6F4E4">Previous or next tab</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+TAB</td>
          <td bgcolor="#F6F4E4">Switching between two recent tabs</td>
          <td bgcolor="#E9E9F3">Ctrl+Shift+F</td>
          <td bgcolor="#F6F4E4">Search in files</td>
        </tr>
      </table>
    </p>

    <br>
    <h3>Editor</h3>
    <p align="center">
      <table border="0" cellspacing="1"
             cellpadding="4" width="95%" align="center">
        <tr>
          <td width="15%" bgcolor="#E9E9F3">Ctrl+Up or Down</td>
          <td width="35%" bgcolor="#F6F4E4">Scrolling up or down without changing cursor position</td>
          <td width="15%" bgcolor="#E9E9F3">Alt+Up or Down</td>
          <td width="35%" bgcolor="#F6F4E4">Move cursor one paragraph up or down</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Alt+Left or Right</td>
          <td bgcolor="#F6F4E4">Move cursor word part left or right</td>
          <td bgcolor="#E9E9F3">Ctrl+Shift+Up or Down</td>
          <td bgcolor="#F6F4E4">Select till the beginning or end of a paragraph</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+Shift+T/M/B</td>
          <td bgcolor="#F6F4E4">Jump to the first position of the first visible line/line in a middle of the visible text/last visible line</td>
          <td bgcolor="#E9E9F3">Ctrl+Z or Ctrl+Y</td>
          <td bgcolor="#F6F4E4">Undo or Redo</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Shift+Del</td>
          <td bgcolor="#F6F4E4">Copy to buffer and delete selected text (if so) or current line</td>
          <td bgcolor="#E9E9F3">Ctrl+= or -</td>
          <td bgcolor="#F6F4E4">Zoom in or out</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+0</td>
          <td bgcolor="#F6F4E4">Reset zoom</td>
          <td bgcolor="#E9E9F3">Ctrl+G</td>
          <td bgcolor="#F6F4E4">Goto line</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+F</td>
          <td bgcolor="#F6F4E4">Initiate incremental search in buffer</td>
          <td bgcolor="#E9E9F3">Ctrl+R</td>
          <td bgcolor="#F6F4E4">Replace in buffer</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">F3 or Shift+F3</td>
          <td bgcolor="#F6F4E4">Search next or previous</td>
          <td bgcolor="#E9E9F3">Ctrl+'</td>
          <td bgcolor="#F6F4E4">Highlight current word</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+, or .</td>
          <td bgcolor="#F6F4E4">Move to the previous or next highlighted word</td>
          <td bgcolor="#E9E9F3">Ctrl+M</td>
          <td bgcolor="#F6F4E4">Comment or uncomment a line or selected lines</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+Space, TAB</td>
          <td bgcolor="#F6F4E4">Code completion</td>
          <td bgcolor="#E9E9F3">Ctrl+F1</td>
          <td bgcolor="#F6F4E4">Context help</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+back slash</td>
          <td bgcolor="#F6F4E4">Goto definition</td>
          <td bgcolor="#E9E9F3">Ctrl+Shift+I</td>
          <td bgcolor="#F6F4E4">Invoke auto indenter</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+]</td>
          <td bgcolor="#F6F4E4">Find occurrences of the current word</td>
          <td bgcolor="#E9E9F3">Alt+U</td>
          <td bgcolor="#F6F4E4">Jump to the beginning of the current function or class</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+forward slash</td>
          <td bgcolor="#F6F4E4">Show or hide a calltip</td>
          <td bgcolor="#E9E9F3">Alt+Shift+cursor keys</td>
          <td bgcolor="#F6F4E4">Rectangular selection</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+mouse selection</td>
          <td bgcolor="#F6F4E4">Rectangular selection</td>
          <td bgcolor="#E9E9F3">Ctrl+B</td>
          <td bgcolor="#F6F4E4">Highlight the current text cursor context in the outline browser without moving the focus</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">F12</td>
          <td bgcolor="#F6F4E4">Make the cursor line the first visible</td>
          <td bgcolor="#E9E9F3">Ctrl+`</td>
          <td bgcolor="#F6F4E4">Select a graphics pane item which corresponds the current text cursor
                                position and scroll the graphics view if needed</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+1</td>
          <td bgcolor="#F6F4E4">Move the focus to the text editor</td>
          <td bgcolor="#E9E9F3">Ctrl+2</td>
          <td bgcolor="#F6F4E4">Move the focus to the flowchart pane</td>
        </tr>
      </table>
    </p>

    <br>
    <h3>Debugger</h3>
    <p align="center">
      <table border="0" cellspacing="1"
             cellpadding="4" width="95%" align="center">
        <tr>
          <td width="15%" bgcolor="#E9E9F3">Shift+F5</td>
          <td width="35%" bgcolor="#F6F4E4">Start debugging the project main script with saved settings</td>
          <td width="15%" bgcolor="#E9E9F3">F5</td>
          <td width="35%" bgcolor="#F6F4E4">Start debugging the current tab script with saved settings</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+Shift+F5</td>
          <td bgcolor="#F6F4E4">Edit debugger settings and start debugging the project main script</td>
          <td bgcolor="#E9E9F3">Ctrl+F5</td>
          <td bgcolor="#F6F4E4">Edit debugger settings and start debugging the current tab script</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+F10</td>
          <td bgcolor="#F6F4E4">Stop the debugging session and kill the i/o console</td>
          <td bgcolor="#E9E9F3">F10</td>
          <td bgcolor="#F6F4E4">Stop the debugging session and keep the i/o console</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">F4</td>
          <td bgcolor="#F6F4E4">Restart the debugging session</td>
          <td bgcolor="#E9E9F3">F6</td>
          <td bgcolor="#F6F4E4">Continue</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">F7</td>
          <td bgcolor="#F6F4E4">Step in</td>
          <td bgcolor="#E9E9F3">F8</td>
          <td bgcolor="#F6F4E4">Step over</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">F9</td>
          <td bgcolor="#F6F4E4">Step out</td>
          <td bgcolor="#E9E9F3">Shift+F6</td>
          <td bgcolor="#F6F4E4">Run to cursor</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+W</td>
          <td bgcolor="#F6F4E4">Show the current debugger line</td>
          <td bgcolor="#E9E9F3"></td>
          <td bgcolor="#F6F4E4"></td>
        </tr>
      </table>
    </p>

    <br>
    <h3>Graphics Pane</h3>
    <p align="center">
      <table border="0" cellspacing="1"
             cellpadding="4" width="95%" align="center">
        <tr>
          <td width="15%" bgcolor="#E9E9F3">Escape</td>
          <td width="35%" bgcolor="#F6F4E4">Clear selection</td>
          <td width="15%" bgcolor="#E9E9F3">Ctrl+`</td>
          <td width="35%" bgcolor="#F6F4E4">Select the first visible item,
                                            move focus to the text editor and
                                            scroll to the line which corresponds to the selected item</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Home</td>
          <td bgcolor="#F6F4E4">Scroll horizontally to the most left position</td>
          <td bgcolor="#E9E9F3">End</td>
          <td bgcolor="#F6F4E4">Scroll horizontally to the most right position</td>
        </tr>
        <tr>
          <td bgcolor="#E9E9F3">Ctrl+Home</td>
          <td bgcolor="#F6F4E4">Scroll to the top left position</td>
          <td bgcolor="#E9E9F3">Ctrl+End</td>
          <td bgcolor="#F6F4E4">Scroll to the bottom left position</td>
        </tr>
      </table>
    </p>

    <br>
    <p align="left">
        Note: The industry common keys are not shown above. Please refer to
        <a href="http://codimension.org/documentation/cheatsheet.html">
           http://codimension.org/documentation/cheatsheet.html</a> for the complete list of
           bindings.
    </p>

    <br><br>""" + self.generateCMLHelp() + """
</div>
</body>
</html>""")

        self.setFileName("")
        self.setShortName("Quick help")

    @staticmethod
    def generateCMLHelp():
        """Generates CML comments help"""
        cmlHelpHeader = """<h3 align="center">CML Comments Reference<br>
                        (CML version """ + str(CMLVersion.VERSION) + ")</h3>"

        cmlCodes = list(CMLVersion.COMMENT_TYPES.keys())
        cmlCodes.sort()

        rows = ['<tr><td width="15%" bgcolor="#E9E9F3"></td>'
                '<td width="85%" bgcolor="#F6F4E4">' +
                CMLVersion.description().replace('\n', '<br/>') +
                '</td></tr>']
        for code in cmlCodes:
            descr = CMLVersion.COMMENT_TYPES[code].description()
            descr = descr.replace("\n", "<br/>")
            rows.append("""<tr><td width="15%" bgcolor="#E9E9F3">""" + code +
                        """</td><td width="85%" bgcolor="#F6F4E4">""" + descr +
                        "</td></tr>")

        return cmlHelpHeader + \
               """<p align="center">
                  <table border="0" cellspacing="1"
                         cellpadding="4" width="95%" align="center">""" + \
               "\n".join(rows) + \
               """</table></p>"""

    def setFocus(self):
        """Sets the focus to the nested html displaying editor"""
        TextTabWidget.setFocus(self)



#
# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2012  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Profiler test"""
# import sys
import time

x = 0
while False:
    time.sleep(0.1)
    x += 1


v = range(175)
z = (x*x for x in range(10))
c = compile('sum([1, 2, 3])', '', 'single')
e = Ellipsis

raise Exception('dkdkdkd')

try:
    raise Exception('hey')
except Exception as exc:
    import sys
    x1, x2, x3 = sys.exc_info()

class A:
    def __init__(self):
        self.__d = 10
        self.d = 20
    def f():
        pass
    @property
    def x(self):
        return self.__x

def f( bla ):
    " F function docstring "
    if bla == -1:
        return -1
#    if bla >= 5:
#        return -1
    return f( bla + 1 )

def g( foo ):
    " g function doc "
    f( foo )

a = A()
b = set()
c = {1: '1', 2: '2'}

f( 0 )
g( -1 )




# -*- coding: utf-8 -*-
#
# Codimension - Python 3 experimental IDE
# Copyright (C) 2010-2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Setup script for the Codimension IDE"""

import os.path
import sys
import io
from setuptools import setup


def getVersion():
    """The version is coming from a file in the source tree"""
    verFileName = 'codimension/cdmverspec.py'
    if not os.path.exists(verFileName):
        print('Cannot find the IDE version file. Expected here: ' +
              verFileName, file=sys.stderr)
        sys.exit(1)

    with open(verFileName) as version_file:
        for line in version_file:
            line = line.strip()
            if line.startswith('version'):
                return line.split('=')[1].strip()[1:-1]
    print('Cannot find a version line in the ' + verFileName,
          file=sys.stderr)
    sys.exit(1)


def getDescription():
    """Provides a short description"""
    return 'Experimental Python 3 IDE which aims at both textual and ' \
           'graphical representation of a code. The graphics is ' \
           'automatically re-generated while the code is typed'


def getLongDescription():
    """Provides the long description"""
    try:
        import pypandoc
        converted = pypandoc.convert('README.md', 'rst').splitlines()
        no_travis = [line for line in converted if 'travis-ci.org' not in line]
        long_description = '\n'.join(no_travis)

        # Pypi index does not like this link
        long_description = long_description.replace('|Build Status|', '')
    except Exception as exc:
        print('pypandoc package is not installed: the markdown '
              'README.md convertion to rst failed: ' + str(exc), file=sys.stderr)
        # pandoc is not installed, fallback to using raw contents
        with io.open('README.md', encoding='utf-8') as f:
            long_description = f.read()

    return long_description


def getRequirements():
    """Provides the requirements list"""
    if not os.path.exists('requirements.txt'):
        print('Could not find requirements.txt', file=sys.stderr)
        sys.exit(1)

    with open('requirements.txt') as f:
        required = f.read().splitlines()
    return required


def getDataFiles():
    """Provides the data files"""
    result = [('share/applications', ['resources/codimension.desktop']),
              ('share/pixmaps', ['resources/codimension.png']),
              ('share/metainfo', ['resources/codimension.appdata.xml'])]
    return result


def getPackageData():
    """Provides the data files"""
    extensions = ['.png', '.svg', '.svgz', '.json', '.css', '.md', '.jar',
                  'README', 'COPYING']
    package_data = [('codimension.pixmaps',
                     'codimension/pixmaps/'),
                    ('codimension.skins',
                     'codimension/skins/'),
                    ('doc',
                     'doc/'),
                    ('doc.md',
                     'doc/md'),
                    ('doc.cml',
                     'doc/cml'),
                    ('doc.plugins',
                     'doc/plugins'),
                    ('doc.technology',
                     'doc/technology'),
                    ('doc.smartzoom',
                     'doc/smartzoom'),
                    ('doc.grouping',
                     'doc/grouping'),
                    ('doc.deadcode',
                     'doc/deadcode'),
                    ('doc.complexity',
                     'doc/complexity'),
                    ('doc.pyflakes',
                     'doc/pyflakes'),
                    ('doc.disassembling',
                     'doc/disassembling'),
                    ('doc.colorschemes',
                     'doc/colorschemes'),
                    ('doc.editorsettings',
                     'doc/editorsettings'),
                    ('doc.dependencies',
                     'doc/dependencies'),
                    ('doc.project',
                     'doc/project'),
                    ('plantuml',
                     'plantuml/')]

    # If a skin needs to be added, then the following item should be also
    # appended:
    # package_data.append(('codimension.skins.myskin',
    #                      'codimension/skins/myskin/'))

    result = {}
    for item in package_data:
        package = item[0]
        matchFiles = []
        for fName in os.listdir(item[1]):
            for ext in extensions:
                if fName.endswith(ext):
                    matchFiles.append(fName)
                    break
        if matchFiles:
            result[package] = matchFiles
    return result


def getPackages():
    """Provides packages"""
    return ['codimension',
            'codimension.analysis',
            'codimension.autocomplete',
            'codimension.diagram',
            'codimension.editor',
            'codimension.flowui',
            'codimension.profiling',
            'codimension.ui',
            'codimension.utils',
            'codimension.debugger', 'codimension.debugger.client',
            'codimension.plugins',
            'codimension.plugins.categories',
            'codimension.plugins.manager',
            'codimension.plugins.vcssupport',
            'codimension.pixmaps',
            'codimension.skins',
            'doc', 'doc.cml', 'doc.plugins', 'doc.technology', 'doc.md',
            'doc.smartzoom', 'doc.grouping', 'doc.deadcode',
            'doc.complexity', 'doc.pyflakes', 'doc.disassembling',
            'doc.colorschemes', 'doc.editorsettings',
            'doc.dependencies', 'doc.project',
            'plantuml']
    # If a myskin skin is to be added as well, then one more package should
    # be mentioned: ..., 'codimension.skins.myskin']


# install_requires=['pypandoc'] could be added but really it needs to only
# at the time of submitting a package to Pypi so it is excluded from the
# dependencies
setup(name='codimension',
      description=getDescription(),
      python_requires='>=3.5, <3.8',
      long_description=getLongDescription(),
      # long_description_content_type does not really work so far
      # long_description_content_type='text/markdown',
      version=getVersion(),
      author='Sergey Satskiy',
      author_email='sergey.satskiy@gmail.com',
      url='https://github.com/SergeySatskiy/codimension',
      license='GPLv3',
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: GNU General Public License (GPL)',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python :: 3'],
      platforms=['any'],
      packages=getPackages(),
      package_data=getPackageData(),
      install_requires=getRequirements(),
      data_files=getDataFiles(),
      entry_points={'gui_scripts':
                    ['codimension = codimension.codimension:main']})



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
The watcher cares of the given directories structures.
If somethig has been changed then it emits a signal.
It accepts a set of exclude filters and a set of dirs to watch.
The watcher will ignore the directories which do not exist.
"""

# pylint: disable=W0702
# pylint: disable=W0703

import os
import os.path
import re
from ui.qt import QObject, QFileSystemWatcher, pyqtSignal


class Watcher(QObject):

    """Filesystem watcher implementation"""

    sigFSChanged = pyqtSignal(list)

    def __init__(self, excludeFilters, dirToWatch):
        QObject.__init__(self)
        self.__dirWatcher = QFileSystemWatcher(self)

        # data members
        self.__excludeFilter = []       # Files exclude filter
        self.__srcDirsToWatch = set()   # Came from the user

        self.__fsTopLevelSnapshot = {}  # Current snapshot
        self.__fsSnapshot = {}          # Current snapshot

        # Sets of dirs which are currently watched
        self.__dirsToWatch = set()
        self.__topLevelDirsToWatch = set()      # Generated till root

        # precompile filters
        for flt in excludeFilters:
            self.__excludeFilter.append(re.compile(flt))

        # Initialise the list of dirs to watch
        self.__srcDirsToWatch.add(dirToWatch)

        self.__topLevelDirsToWatch = self.__buildTopDirsList(
            self.__srcDirsToWatch)
        self.__fsTopLevelSnapshot = self.__buildTopLevelSnapshot(
            self.__topLevelDirsToWatch, self.__srcDirsToWatch)
        self.__dirsToWatch = self.__buildSnapshot()

        # Here __dirsToWatch and __topLevelDirsToWatch have a complete
        # set of what should be watched

        # Add the dirs to the watcher
        dirs = []
        for path in self.__dirsToWatch | self.__topLevelDirsToWatch:
            dirs.append(path)
        self.__dirWatcher.addPaths(dirs)
        self.__dirWatcher.directoryChanged.connect(self.__onDirChanged)

        # self.debug()

    @staticmethod
    def __buildTopDirsList(srcDirs):
        """Takes a list of dirs to be watched and builds top dirs set"""
        topDirsList = set()
        for path in srcDirs:
            parts = path.split(os.path.sep)
            for index in range(1, len(parts) - 1):
                candidate = os.path.sep.join(parts[0:index]) + os.path.sep
                if os.path.exists(candidate):
                    if os.access(candidate, os.R_OK):
                        topDirsList.add(candidate)
        return topDirsList

    @staticmethod
    def __buildTopLevelSnapshot(topLevelDirs, srcDirs):
        """Takes top level dirs and builds their snapshot"""
        snapshot = {}
        for path in topLevelDirs:
            itemsSet = set()
            # search for all the dirs to be watched
            for candidate in topLevelDirs | srcDirs:
                if len(candidate) <= len(path):
                    continue
                if candidate.startswith(path):
                    candidate = candidate[len(path):]
                    slashIndex = candidate.find(os.path.sep) + 1
                    item = candidate[:slashIndex]
                    if os.path.exists(path + item):
                        itemsSet.add(item)
            snapshot[path] = itemsSet
        return snapshot

    def __buildSnapshot(self):
        """Builds the filesystem snapshot"""
        snapshotDirs = set()
        for path in self.__srcDirsToWatch:
            self.__addSnapshotPath(path, snapshotDirs)
        return snapshotDirs

    def __addSnapshotPath(self, path, snapshotDirs, itemsToReport=None):
        """Adds one path to the FS snapshot"""
        if not os.path.exists(path):
            return

        snapshotDirs.add(path)
        dirItems = set()
        for item in os.listdir(path):
            if self.__shouldExclude(item):
                continue
            if os.path.isdir(path + item):
                dirName = path + item + os.path.sep
                dirItems.add(item + os.path.sep)
                if itemsToReport is not None:
                    itemsToReport.append("+" + dirName)
                self.__addSnapshotPath(dirName, snapshotDirs, itemsToReport)
                continue
            dirItems.add(item)
            if itemsToReport is not None:
                itemsToReport.append("+" + path + item)
        self.__fsSnapshot[path] = dirItems
        return

    def __onDirChanged(self, path):
        """Triggered when the dir is changed"""
        if not path.endswith(os.path.sep):
            path = path + os.path.sep

        # Check if it is a top level dir
        try:
            oldSet = self.__fsTopLevelSnapshot[path]

            # Build a new set of what is in that top level dir
            newSet = set()
            for item in os.listdir(path):
                if not os.path.isdir(path + item):
                    continue    # Only dirs are of interest for the top level
                item = item + os.path.sep
                if item in oldSet:
                    newSet.add(item)
            # Now we have an old set and a new one with those from the old
            # which actually exist
            diff = oldSet - newSet

            # diff are those which disappeared. We need to do the following:
            # - build a list of all the items in the fs snapshot which start
            #   from this dir
            # - build a list of dirs which should be deregistered from the
            #   watcher. This list includes both top level and project level
            # - deregister dirs from the watcher
            # - emit a signal of what disappeared
            if not diff:
                return  # no changes

            self.__fsTopLevelSnapshot[path] = newSet

            dirsToBeRemoved = []
            itemsToReport = []

            for item in diff:
                self.__processRemoveTopDir(path + item, dirsToBeRemoved,
                                           itemsToReport)

            # Here: it is possible that the last dir to watch disappeared
            if not newSet:
                # There is nothing to watch here anymore
                dirsToBeRemoved.append(path)
                del self.__fsTopLevelSnapshot[path]

                parts = path[1:-1].split(os.path.sep)
                for index in range(len(parts) - 2, 0, -1):
                    candidate = os.path.sep + \
                                os.path.sep.join(parts[0:index]) + \
                                os.path.sep
                    dirSet = self.__fsTopLevelSnapshot[candidate]
                    dirSet.remove(parts[index + 1] + os.path.sep)
                    if not dirSet:
                        dirsToBeRemoved.append(candidate)
                        del self.__fsTopLevelSnapshot[candidate]
                        continue
                    break   # it is not the last item in the set

            # Update the watcher
            if dirsToBeRemoved:
                self.__dirWatcher.removePaths(dirsToBeRemoved)

            # Report
            if itemsToReport:
                self.sigFSChanged.emit(itemsToReport)
            return
        except:
            # it is not a top level dir - no key
            pass

        # Here: the change is in the project level dir
        try:
            oldSet = self.__fsSnapshot[path]

            # Build a new set of what is in that top level dir
            newSet = set()
            for item in os.listdir(path):
                if self.__shouldExclude(item):
                    continue
                if os.path.isdir(path + item):
                    newSet.add(item + os.path.sep)
                else:
                    newSet.add(item)

            # Here: we have a new and old snapshots
            # Lets calculate the difference
            deletedItems = oldSet - newSet
            addedItems = newSet - oldSet

            if not deletedItems and not addedItems:
                return  # No changes

            # Update the changed dir set
            self.__fsSnapshot[path] = newSet

            # We need to build some lists:
            # - list of files which were added
            # - list of dirs which were added
            # - list of files which were deleted
            # - list of dirs which were deleted
            # The deleted dirs must be unregistered in the watcher
            # The added dirs must be registered
            itemsToReport = []
            dirsToBeAdded = []
            dirsToBeRemoved = []

            for item in addedItems:
                if item.endswith(os.path.sep):
                    # directory was added
                    self.__processAddedDir(path + item,
                                           dirsToBeAdded, itemsToReport)
                else:
                    itemsToReport.append("+" + path + item)

            for item in deletedItems:
                if item.endswith(os.path.sep):
                    # directory was deleted
                    self.__processRemovedDir(path + item,
                                             dirsToBeRemoved, itemsToReport)
                else:
                    itemsToReport.append("-" + path + item)

            # Update the watcher
            if dirsToBeRemoved:
                self.__dirWatcher.removePaths(dirsToBeRemoved)
            if dirsToBeAdded:
                self.__dirWatcher.addPaths(dirsToBeAdded)

            # Report
            self.sigFSChanged.emit(itemsToReport)
        except:
            # It could be a queued signal about what was already reported
            pass

        # self.debug()
        return

    def __shouldExclude(self, name):
        """Tests if a file must be excluded"""
        for excl in self.__excludeFilter:
            if excl.match(name):
                return True
        return False

    def __processAddedDir(self, path, dirsToBeAdded, itemsToReport):
        """called for an appeared dir in the project tree"""
        dirsToBeAdded.append(path)
        itemsToReport.append("+" + path)

        # it should add dirs recursively into the snapshot and care
        # of the items to report
        dirItems = set()
        for item in os.listdir(path):
            if self.__shouldExclude(item):
                continue
            if os.path.isdir(path + item):
                dirName = path + item + os.path.sep
                dirItems.add(item + os.path.sep)
                self.__processAddedDir(dirName, dirsToBeAdded, itemsToReport)
                continue
            itemsToReport.append("+" + path + item)
            dirItems.add(item)
        self.__fsSnapshot[path] = dirItems

    def __processRemovedDir(self, path, dirsToBeRemoved, itemsToReport):
        """called for a disappeared dir in the project tree"""
        # it should remove the dirs recursively from the fs snapshot
        # and care of items to report
        dirsToBeRemoved.append(path)
        itemsToReport.append("-" + path)

        oldSet = self.__fsSnapshot[path]
        for item in oldSet:
            if item.endswith(os.path.sep):
                # Nested dir
                self.__processRemovedDir(path + item, dirsToBeRemoved,
                                         itemsToReport)
            else:
                # a file
                itemsToReport.append("-" + path + item)
        del self.__fsSnapshot[path]

    def __processRemoveTopDir(self, path, dirsToBeRemoved, itemsToReport):
        """Called for a disappeared top level dir"""
        if path in self.__fsTopLevelSnapshot:
            # It is still a top level dir
            dirsToBeRemoved.append(path)
            for item in self.__fsTopLevelSnapshot[path]:
                self.__processRemoveTopDir(path + item, dirsToBeRemoved,
                                           itemsToReport)
            del self.__fsTopLevelSnapshot[path]
        else:
            # This is a project level dir
            self.__processRemovedDir(path, dirsToBeRemoved,
                                     itemsToReport)

    def reset(self):
        """Resets the watcher (it does not report any changes)"""
        self.__dirWatcher.removePaths(self.__dirWatcher.directories())

        self.__srcDirsToWatch = set()

        self.__fsTopLevelSnapshot = {}
        self.__fsSnapshot = {}

        self.__dirsToWatch = set()
        self.__topLevelDirsToWatch = set()

    def registerDir(self, path):
        """Adds a directory to the list of watched ones"""
        if not path.endswith(os.path.sep):
            path = path + os.path.sep

        if path in self.__srcDirsToWatch:
            return  # It is there already

        # It is necessary to do the following:
        # - add the dir to the fs snapshot
        # - collect dirs to add to the watcher
        # - collect items to report
        self.__srcDirsToWatch.add(path)

        dirsToWatch = set()
        itemsToReport = []
        self.__registerDir(path, dirsToWatch, itemsToReport)

        # It might be that top level dirs should be updated too
        newTopLevelDirsToWatch = self.__buildTopDirsList(self.__srcDirsToWatch)
        addedDirs = newTopLevelDirsToWatch - self.__topLevelDirsToWatch

        for item in addedDirs:
            dirsToWatch.add(item)

            # Identify items to be watched by this dir
            dirItems = set()
            for candidate in newTopLevelDirsToWatch | self.__srcDirsToWatch:
                if len(candidate) <= len(item):
                    continue
                if candidate.startswith(item):
                    candidate = candidate[len(item):]
                    slashIndex = candidate.find(os.path.sep) + 1
                    dirName = candidate[:slashIndex]
                    if os.path.exists(item + dirName):
                        dirItems.add(dirName)
            # Update the top level dirs snapshot
            self.__fsTopLevelSnapshot[item] = dirItems

        # Update the top level snapshot with the added dir
        upperDir = os.path.dirname(path[:-1]) + os.path.sep
        dirName = path.replace(upperDir, '')
        self.__fsTopLevelSnapshot[upperDir].add(dirName)

        # Update the list of top level dirs to watch
        self.__topLevelDirsToWatch = newTopLevelDirsToWatch

        # Update the watcher
        if dirsToWatch:
            dirs = []
            for item in dirsToWatch:
                dirs.append(item)
            self.__dirWatcher.addPaths(dirs)

        # Report the changes
        if itemsToReport:
            self.sigFSChanged.emit(itemsToReport)

        # self.debug()
        return

    def __registerDir(self, path, dirsToWatch, itemsToReport):
        """Adds one path to the FS snapshot"""
        if not os.path.exists(path):
            return

        dirsToWatch.add(path)
        itemsToReport.append("+" + path)

        dirItems = set()
        for item in os.listdir(path):
            if self.__shouldExclude(item):
                continue
            if os.path.isdir(path + item):
                dirName = path + item + os.path.sep
                dirItems.add(item + os.path.sep)
                itemsToReport.append("+" + path + item + os.path.sep)
                self.__addSnapshotPath(dirName, dirsToWatch, itemsToReport)
                continue
            dirItems.add(item)
            itemsToReport.append("+" + path + item)
        self.__fsSnapshot[path] = dirItems

    def deregisterDir(self, path):
        """Removes the directory from the list of the watched ones"""
        if not path.endswith(os.path.sep):
            path = path + os.path.sep

        if path not in self.__srcDirsToWatch:
            return  # It is not there already
        self.__srcDirsToWatch.remove(path)

        # It is necessary to do the following:
        # - remove the dir from the fs snapshot
        # - collect the dirs to be removed from watching
        # - collect item to report

        itemsToReport = []
        dirsToBeRemoved = []

        self.__deregisterDir(path, dirsToBeRemoved, itemsToReport)

        # It is possible that some of the top level watched dirs should be
        # removed as well
        newTopLevelDirsToWatch = self.__buildTopDirsList(self.__srcDirsToWatch)
        deletedDirs = self.__topLevelDirsToWatch - newTopLevelDirsToWatch

        for item in deletedDirs:
            dirsToBeRemoved.append(item)
            del self.__fsTopLevelSnapshot[item]

        # It might be the case that some of the items should be deleted in the
        # top level dirs sets
        for dirName in self.__fsTopLevelSnapshot:
            itemsSet = self.__fsTopLevelSnapshot[dirName]
            for item in itemsSet:
                candidate = dirName + item
                if candidate == path or candidate in deletedDirs:
                    itemsSet.remove(item)
                    self.__fsTopLevelSnapshot[dirName] = itemsSet
                    break

        # Update the list of dirs to be watched
        self.__topLevelDirsToWatch = newTopLevelDirsToWatch

        # Update the watcher
        if dirsToBeRemoved:
            self.__dirWatcher.removePaths(dirsToBeRemoved)

        # Report the changes
        if itemsToReport:
            self.sigFSChanged.emit(itemsToReport)

        # self.debug()

    def __deregisterDir(self, path, dirsToBeRemoved, itemsToReport):
        """Deregisters a directory recursively"""
        dirsToBeRemoved.append(path)
        itemsToReport.append("-" + path)
        if path in self.__fsTopLevelSnapshot:
            # This is a top level dir
            for item in self.__fsTopLevelSnapshot[path]:
                if item.endswith(os.path.sep):
                    # It's a dir
                    self.__deregisterDir(path + item, dirsToBeRemoved,
                                         itemsToReport)
                else:
                    # It's a file
                    itemsToReport.append("-" + path + item)
            del self.__fsTopLevelSnapshot[path]
            return

        # It is from an a project level snapshot
        if path in self.__fsSnapshot:
            for item in self.__fsSnapshot[path]:
                if item.endswith(os.path.sep):
                    # It's a dir
                    self.__deregisterDir(path + item, dirsToBeRemoved,
                                         itemsToReport)
                else:
                    # It's a file
                    itemsToReport.append("-" + path + item)
            del self.__fsSnapshot[path]
        return

    def debug(self):
        """Debugging printouts"""
        print("Top level dirs to watch: " + str(self.__topLevelDirsToWatch))
        print("Project dirs to watch: " + str(self.__dirsToWatch))

        print("Top level snapshot: " + str(self.__fsTopLevelSnapshot))
        print("Project snapshot: " + str(self.__fsSnapshot))
